# This function includes various function used for general functionalities such as plotting, grid search

import copy
from typing import List, Tuple

from matplotlib.collections import PatchCollection
from typing_extensions import deprecated

import geojson
import math
from math import isinf
import heapq
import pandas as pd
import numpy as np
import networkx as nx
import logging as log

from pandapower.auxiliary import pandapowerNet
from pandapower.topology.create_graph import create_nxgraph
from pandapower.create import create_bus, create_line_from_parameters
from pandapower.plotting.collections import create_annotation_collection, create_line_collection, \
    create_bus_collection, create_line_switch_collection, draw_collections, create_trafo_collection, \
    create_trafo_connection_collection, create_load_collection, create_bus_bus_switch_collection
from pandapower.toolbox.grid_modification import fuse_buses
from pandapower.toolbox.element_selection import get_connected_buses_at_element, get_connected_elements, next_bus
from pandapower.run import runpp
from pandapower.shortcircuit.calc_sc import calc_sc

import warnings

logger = log.getLogger(__name__)

try:
    import matplotlib.pyplot as plt

    MATPLOTLIB_INSTALLED = True
except ImportError:
    MATPLOTLIB_INSTALLED = False

try:
    import mplcursors

    MPLCURSORS_INSTALLED = True
except ImportError:
    MPLCURSORS_INSTALLED = False
    logger.info('could not import mplcursors')

warnings.filterwarnings('ignore')


def _get_coords_from_bus_idx(net: pandapowerNet, bus_idx: pd.Index) -> List[Tuple[float, float]]:
    try:
        bl = net.bus.dropna(subset=["geo"]).loc[bus_idx, 'geo']
        if isinstance(bl, pd.Series):
            return bl.apply(geojson.loads).apply(geojson.utils.coords).apply(next).to_list()
        else:
            return [next(geojson.utils.coords(geojson.loads(bl)))]
    except KeyError:
        logger.error(f"Bus {bus_idx} not found in net.bus.geo")
    return []


def _get_coords_from_line_idx(net: pandapowerNet, line_idx: pd.Index) -> List[Tuple[float, float]]:
    try:
        ll = net.line.dropna(subset=["geo"]).loc[line_idx, 'geo']
        if isinstance(ll, pd.Series):
            return ll.apply(geojson.loads).apply(geojson.utils.coords).apply(next).to_list()
        else:
            return [next(geojson.utils.coords(geojson.loads(ll)))]
    except KeyError:
        logger.error(f"Line {line_idx} not found in net.line.geo")
    return []


def create_sc_bus(net_copy, sc_line_id, sc_fraction):
    # This function creates a short-circuit location (a bus) on a line.
    net = copy.deepcopy(net_copy)
    if sc_fraction < 0 or sc_fraction > 1:
        print("Please select line location that is between 0 and 1")
        return

    max_idx_line = max(net.line.index)
    max_idx_bus = max(net.bus.index)

    aux_line = net.line.loc[sc_line_id]

    # get bus voltage of the short circuit bus
    bus_vn_kv = net.bus.vn_kv.at[aux_line.from_bus]

    # set new virtual short circuit bus with give line and location
    bus_sc = create_bus(net, name="Bus_SC", vn_kv=bus_vn_kv, type="b", index=max_idx_bus + 1)

    # sim bench grids
    if 's_sc_max_mva' not in net.ext_grid:
        print('input s_sc_max_mva or taking 1000')
        net.ext_grid['s_sc_max_mva'] = 1000
    if 'rx_max' not in net.ext_grid:
        print('input rx_max or taking 0.1')
        net.ext_grid['rx_max'] = 0.1
    if 'k' not in net.sgen and len(net.sgen) != 0:
        print('input  Ratio of nominal current to short circuit current- k or  taking k=1')
        net.sgen['k'] = 1

    # set new lines
    sc_line1 = sc_line_id
    net.line.at[sc_line1, 'to_bus'] = bus_sc
    net.line.at[sc_line1, 'length_km'] *= sc_fraction

    sc_line2 = create_line_from_parameters(net, bus_sc, aux_line.to_bus,
                                              length_km=aux_line.length_km * (1 - sc_fraction),
                                              index=max_idx_line + 1, r_ohm_per_km=aux_line.r_ohm_per_km,
                                              x_ohm_per_km=aux_line.x_ohm_per_km, c_nf_per_km=aux_line.c_nf_per_km,
                                              max_i_ka=aux_line.max_i_ka)

    if 'endtemp_degree' in net.line.columns:
        net.line.at[sc_line2, "endtemp_degree"] = net.line.endtemp_degree.at[sc_line1]

    net.line = net.line.sort_index()

    # check if switches are connected to the line and set the switches to new lines
    for switch_id in net.switch.index:
        if (aux_line.from_bus == net.switch.bus[switch_id]) & (net.switch.element[switch_id] == sc_line_id):
            net.switch.loc[switch_id, 'element'] = sc_line1
        elif (aux_line.to_bus == net.switch.bus[switch_id]) & (net.switch.element[switch_id] == sc_line_id):
            net.switch.element[switch_id] = sc_line2

    # set geodata for new bus
    net.bus.loc[max_idx_bus + 1, 'geo'] = None

    x1, y1 = _get_coords_from_bus_idx(net, aux_line.from_bus)[0]
    x2, y2 = _get_coords_from_bus_idx(net, aux_line.to_bus)[0]

    net.bus.geo.at[max_idx_bus + 1] = geojson.dumps(
        geojson.Point((sc_fraction * (x2 - x1) + x1, sc_fraction * (y2 - y1) + y1)), sort_keys=True)
    return net


def calc_faults_at_full_line(net, line, location_step_size=0.01, start_location=0.01, end_location=1, sc_case="min"):
    # functon to create sc at full line
    i = 0

    fault_currents = []
    max_bus_idx = max(net.bus.index)
    location = start_location
    while location < end_location:
        net_sc = create_sc_bus(net, line, location)
        calc_sc(net_sc, case=sc_case)

        fault_currents.append(net_sc.res_bus_sc.ikss_ka.at[max_bus_idx + 1])

        i += 1
        location = start_location + i * location_step_size
    return fault_currents


def get_opposite_side_bus_from_switch(net, switch_id):
    # get the frm and to bus of switch
    line_idx = net.switch.element.at[switch_id]
    is_from_bus = get_from_bus_info_switch(net, switch_id)

    if is_from_bus:
        opp_bus_idx = net.line.to_bus.at[line_idx]
    else:
        opp_bus_idx = net.line.from_bus.at[line_idx]

    return opp_bus_idx


def get_opposite_side_bus_from_bus_line(net, bus_idx, line_idx):
    # get the from abd to bus of given line
    is_from_bus = get_from_bus_info_bus_line(net, bus_idx, line_idx)

    if is_from_bus:
        opp_bus_idx = net.line.to_bus.at[line_idx]
    else:
        opp_bus_idx = net.line.from_bus.at[line_idx]

    return opp_bus_idx


def get_from_bus_info_switch(net, switch_id):
    # get the from bus of given switch id
    bus_idx = net.switch.bus.at[switch_id]
    line_idx = net.switch.element.at[switch_id]

    return bus_idx == net.line.from_bus.at[line_idx]


def get_from_bus_info_bus_line(net, bus_idx, line_idx):
    # get bus nfo of given line
    return bus_idx == net.line.from_bus.at[line_idx]


def get_line_impedance(net, line_idx):
    # get line impedence
    line_length = net.line.length_km.at[line_idx]
    line_r_per_km = net.line.r_ohm_per_km.at[line_idx]
    line_x_per_km = net.line.x_ohm_per_km.at[line_idx]
    z_line = complex(line_r_per_km * line_length, line_x_per_km * line_length)  # Z = R + jX
    return z_line


def get_lowest_impedance_line(net: pandapowerNet, lines):
    # get the low impedenceline
    min_imp_line = None
    min_impedance = float('inf')
    for line in lines:
        impedance = abs(get_line_impedance(net, line))
        if impedance < min_impedance:
            min_impedance = impedance
            min_imp_line = line
    return min_imp_line


def check_for_closed_bus_switches(net_copy):
    # closed switches
    net = copy.deepcopy(net_copy)
    closed_bus_switches = net.switch.loc[(net.switch.et == "b") & (net.switch.closed == True)]

    if len(closed_bus_switches) > 0:
        net = fuse_bus_switches(net, closed_bus_switches)

    return net


def fuse_bus_switches(net, bus_switches):
    # get fused switches
    for bus_switch in bus_switches.index:
        bus1 = net.switch.bus.at[bus_switch]
        bus2 = net.switch.element.at[bus_switch]

        fuse_buses(net, bus1, bus2)

    return net


def get_fault_annotation(net: pandapowerNet, fault_current: float = .0, font_size_bus: float = 0.06) -> PatchCollection:
    max_bus_idx = max(net.bus.dropna(subset=['geo']).index)
    fault_text = f'\tI_sc = {fault_current}kA'

    fault_geo_x_y: Tuple[float, float] = next(geojson.utils.coords(geojson.loads(net.bus.geo.at[max_bus_idx])))
    fault_geo_x_y = (fault_geo_x_y[0], fault_geo_x_y[1] - font_size_bus + 0.02)

    # list of new geo data for line (half position of switch)
    fault_annotate: PatchCollection = create_annotation_collection(
        texts=[fault_text],
        coords=[fault_geo_x_y],
        size=font_size_bus,
        prop=None
    )

    return fault_annotate


def get_sc_location_annotation(net: pandapowerNet, sc_location: float, font_size_bus: float = 0.06) -> PatchCollection:
    max_bus_idx = max(net.bus.dropna(subset=['geo']).index)
    sc_text = f'\tsc_location: {sc_location * 100}%'

    # list of new geo data for line (middle of  position of switch)
    sc_geo_x_y = next(geojson.utils.coords(geojson.loads(net.bus.geo.at[max_bus_idx])))
    sc_geo_x_y = (sc_geo_x_y[0], sc_geo_x_y[1] + 0.02)

    sc_annotate: PatchCollection = create_annotation_collection(
        texts=[sc_text],
        coords=[sc_geo_x_y],
        size=font_size_bus,
        prop=None
    )

    return sc_annotate


def plot_tripped_grid(net, trip_decisions, sc_location, bus_size=0.055, plot_annotations=True):
    # plot the tripped grid of net_sc
    if MPLCURSORS_INSTALLED:
        mplcursors.cursor(hover=False)

    # plot grid and color the according switches - instantaneous tripping red, int backup tripping orange and tripping_time_auto backuo-yellow

    ext_grid_busses = net.ext_grid.bus.values
    fault_location = [max(net.bus.index)]

    lc = create_line_collection(net, lines=net.line.index, zorder=0)

    bc_extgrid = create_bus_collection(net, buses=ext_grid_busses, zorder=1, size=bus_size, patch_type="rect")

    bc = create_bus_collection(net, buses=set(net.bus.index) - set(ext_grid_busses) - set(fault_location),
                                    zorder=2, color="black", size=bus_size)

    bc_fault_location = create_bus_collection(net, buses=set(fault_location), zorder=3, color="red", size=bus_size,
                                                   patch_type="circle")

    collection = [lc, bc_extgrid, bc, bc_fault_location]

    tripping_times = []

    for trip_idx in range(len(trip_decisions)):
        trip_time = trip_decisions[trip_idx].get("Trip time [s]")
        tripping_times.append(trip_time)
    tripping_times = [v for v in tripping_times if not isinf(v)]
    backup_tripping_times = copy.deepcopy(tripping_times)
    backup_tripping_times.remove(min(backup_tripping_times)) and backup_tripping_times.remove(
        heapq.nsmallest(2, backup_tripping_times)[-1])

    inst_trip_switches = []
    backup_trip_switches = []
    inst_backup_switches = []

    for trip_idx in range(len(trip_decisions)):
        trip_decision = trip_decisions[trip_idx]
        switch_id = trip_decision.get("Switch ID")
        trip = trip_decision.get("Trip")
        trip_time = trip_decision.get("Trip time [s]")

        if trip_time == heapq.nsmallest(2, tripping_times)[-1]:
            inst_backup_switches.append(switch_id)

        if trip_time == min(tripping_times):
            inst_trip_switches.append(switch_id)

        if trip_time in backup_tripping_times and trip == True:
            backup_trip_switches.append(switch_id)

    dist_to_bus = bus_size * 3.25

    # Inst relay trip, red colour
    if len(inst_trip_switches) > 0:
        sc_inst = create_line_switch_collection(net, size=bus_size, distance_to_bus=dist_to_bus, color="red",
                                                     switches=inst_trip_switches)
        collection.append(sc_inst)

    # backup relay based on time grade (yellow colour)
    if len(backup_trip_switches) > 0:
        sc_backup = create_line_switch_collection(net, size=bus_size, distance_to_bus=dist_to_bus, color="yellow",
                                                       switches=backup_trip_switches)

        collection.append(sc_backup)

    # orange colour for inst_backup relay
    if len(inst_backup_switches) > 0:
        instant_sc_backup = create_line_switch_collection(net, size=bus_size, distance_to_bus=dist_to_bus,
                                                               color="orange", switches=inst_backup_switches)

        collection.append(instant_sc_backup)

    len_sc = len(set(net.switch.index) - set(inst_trip_switches) - set(backup_trip_switches))

    if len_sc != 0:
        # closed switch-black
        sc = create_line_switch_collection(net, size=bus_size, distance_to_bus=dist_to_bus, color="black",
                                                switches=set(net.switch.index) - set(inst_trip_switches) - set(
                                                    backup_trip_switches))
        collection.append(sc)

    # make annotations optional (if True then annotate else only plot)
    if plot_annotations:
        # line annotations
        line_text = []
        line_geodata = []

        fault_current: float = .0

        # for Switches in trip_decisions:
        for line in net.line.index:

            if line == max(net.line.index):
                break

            # annotate line_id
            text_line = r"  line_" + str(line)  # + ",sw_"+str(Switch_index)

            # get bus_index from the line (from switch)
            get_bus_index = get_connected_buses_at_element(net, element_index=line, element_type='l',
                                                              respect_in_service=False)

            bus_list = list(get_bus_index)
            bus_coords: List[Tuple[float, float]] = [
                geojson.utils.coords(geojson.loads(net.bus.geo.at[bus])) for bus in bus_list
            ]

            # TODO:
            #  place annotations on middle of the line
            line_geo_x = (bus_coords[0][0] + bus_coords[1][0]) / 2
            line_geo_y = ((bus_coords[0][1] + bus_coords[1][1]) / 2) + 0.05

            line_geo_x_y = [line_geo_x, line_geo_y]

            # list of new geo data for line (half position of switch)
            line_geodata.append(tuple(line_geo_x_y))

            fault_current = round(net.res_bus_sc['ikss_ka'].at[max(net.bus.index)], 2)
            # round(Switches['Fault Current'],2)

            line_text.append(text_line)

        # line annotations to collections for plotting
        line_annotate = create_annotation_collection(texts=line_text, coords=line_geodata, size=0.06, prop=None)
        collection.append(line_annotate)

        # Bus Annotatations
        bus_text = []
        for i in net.bus.geo.dropna().index:
            bus_texts = 'bus_' + str(i)

            bus_text.append(bus_texts)

        bus_text = bus_text[:-1]

        bus_geodata = net.bus.geo.dropna().apply(geojson.loads).apply(geojson.utils.coords).apply(next).to_list()

        # placing bus
        bus_index = [(x[0] - 0.11, x[1] + 0.095) for x in bus_geodata]

        # TODO:
        bus_annotate = create_annotation_collection(texts=bus_text, coords=bus_index, size=0.06, prop=None)
        collection.append(bus_annotate)

        # Short circuit annotations
        collection.append(get_fault_annotation(net, fault_current))

        # sc_location annotation
        collection.append(get_sc_location_annotation(net, sc_location))

        # switch annotations
        # from pandapower.protection.implemeutility_functions import switch_geodata
        switch_text = []
        for Switches in trip_decisions:
            Switch_index = Switches['Switch ID']

            text_switch = r"sw_" + str(Switch_index)
            switch_text.append(text_switch)

        switch_geodata = switch_geodatas(net, size=bus_size, distance_to_bus=3.25 * bus_size)
        i = 0
        for i in range(len(switch_geodata)):
            switch_geodata[i]['x'] = switch_geodata[i]['x'] - 0.085  # scale the value if annotations overlap
            switch_geodata[i]['y'] = switch_geodata[i]['y'] + 0.055  # scale the value if annotations overlap
            i = i + 1
        switch_annotate = create_annotation_collection(texts=switch_text, coords=switch_geodata, size=0.06,
                                                            prop=None)
        collection.append(switch_annotate)
    draw_collections(collection)


def plot_tripped_grid_protection_device(net, trip_decisions, sc_location, sc_bus, bus_size=0.055,
                                        plot_annotations=True):
    # plot the tripped grid of net_sc with networks using ProtectionDevice class
    if MPLCURSORS_INSTALLED:
        mplcursors.cursor(hover=False)

    # plot grid and color the according switches - instantaneous tripping red, int backup tripping orange and tripping_time_auto backuo-yellow

    ext_grid_busses = net.ext_grid.bus.values
    fault_location = [max(net.bus.index)]

    lc = create_line_collection(net, lines=net.line.index, zorder=0)

    bc_extgrid = create_bus_collection(net, buses=ext_grid_busses, zorder=1, size=bus_size, patch_type="rect")

    bc = create_bus_collection(net, buses=set(net.bus.index) - set(ext_grid_busses) - set(fault_location),
                                    zorder=2, color="black", size=bus_size)

    bc_fault_location = create_bus_collection(net, buses=set(fault_location), zorder=3, color="red", size=bus_size,
                                                   patch_type="circle")

    collection = [lc, bc_extgrid, bc, bc_fault_location]

    tripping_times = []

    for trip_idx in range(len(trip_decisions)):
        trip_time = trip_decisions.trip_melt_time_s.at[trip_idx]
        tripping_times.append(trip_time)
    tripping_times = [v for v in tripping_times if not isinf(v)]
    if len(tripping_times) == 0:
        return
    backup_tripping_times = copy.deepcopy(tripping_times)
    backup_tripping_times.remove(min(backup_tripping_times)) and backup_tripping_times.remove(
        heapq.nsmallest(2, backup_tripping_times)[-1])

    inst_trip_switches = []
    backup_trip_switches = []
    inst_backup_switches = []
    bus_bus_switches = net.switch.index[net.switch.et == "b"]

    # add trafo to collection
    if len(net.trafo) > 0:
        trafo_collection = create_trafo_collection(net, size=2 * bus_size)
        trafo_conn_collection = create_trafo_connection_collection(net)
        collection.append(trafo_collection)
        collection.append(trafo_conn_collection)
    # add load to collection
    if len(net.load) > 0:
        load_collection = create_load_collection(net, size=2 * bus_size)
        collection.append(load_collection)

    for trip_idx in range(len(trip_decisions)):
        trip_decision = trip_decisions.iloc[[trip_idx]]
        switch_id = trip_decision.switch_id.at[trip_idx]
        trip = trip_decision.trip_melt.at[trip_idx]
        trip_time = trip_decision.trip_melt_time_s.at[trip_idx]

        if trip_time == heapq.nsmallest(2, tripping_times)[-1]:
            inst_backup_switches.append(switch_id)

        if trip_time == min(tripping_times):
            inst_trip_switches.append(switch_id)

        if trip_time in backup_tripping_times and trip == True:
            backup_trip_switches.append(switch_id)

    dist_to_bus = bus_size * 3.25

    # Inst relay trip, red colour
    if len(inst_trip_switches) > 0:
        sc_inst = create_line_switch_collection(net, size=bus_size, distance_to_bus=dist_to_bus, color="red",
                                                     switches=set(inst_trip_switches) - set(bus_bus_switches))
        bb_inst = create_bus_bus_switch_collection(net, size=bus_size, helper_line_color="red")
        collection.append(sc_inst)
        collection.append(bb_inst)

    # backup relay based on time grade (yellow colour)
    if len(backup_trip_switches) > 0:
        sc_backup = create_line_switch_collection(net, size=bus_size, distance_to_bus=dist_to_bus, color="yellow",
                                                       switches=backup_trip_switches)
        bb_backup = create_bus_bus_switch_collection(net, size=bus_size, helper_line_color="yellow")
        collection.append(sc_backup)
        collection.append(bb_backup)

    # orange colour for inst_backup relay
    if len(inst_backup_switches) > 0:
        instant_sc_backup = create_line_switch_collection(net, size=bus_size, distance_to_bus=dist_to_bus,
                                                               color="orange", switches=inst_backup_switches)
        instant_bb_backup = create_bus_bus_switch_collection(net, size=bus_size, helper_line_color="orange")
        collection.append(instant_sc_backup)

    len_sc = len(set(net.switch.index) - set(inst_trip_switches) - set(backup_trip_switches))

    if len_sc != 0:
        # closed switch-black
        sc = create_line_switch_collection(net, size=bus_size, distance_to_bus=dist_to_bus, color="black",
                                                switches=set(net.switch.index) - set(inst_trip_switches) - set(
                                                    backup_trip_switches) - set(bus_bus_switches))
        bb = create_bus_bus_switch_collection(net, size=bus_size, helper_line_color="black")

        collection.append(sc)
        collection.append(bb)

    # make annotations optional (if True then annotate else only plot)
    if plot_annotations:
        # line annotations
        line_text = []
        line_geodata = []

        fault_current = None

        # for Switches in trip_decisions:
        for line in net.line.index:

            if line == max(net.line.index):
                break

            # annotate line_id
            text_line = r"  line_" + str(line)  # + ",sw_"+str(Switch_index)

            # get bus_index from the line (from switch)
            get_bus_index = get_connected_buses_at_element(net, element_index=line, element_type='l',
                                                              respect_in_service=False)

            bus_list = list(get_bus_index)

            # place annotations on middle of the line
            bus_coords = list(
                zip(*net.bus.geo.iloc[bus_list[0:2]].apply(geojson.loads).apply(geojson.utils.coords).apply(
                    next).to_list()))
            line_geo_x_y = [sum(x) / 2 for x in bus_coords]
            line_geo_x_y[1] += 0.05

            # list of new geo data for line (half position of switch)
            line_geodata.append(tuple(line_geo_x_y))

            fault_current = round(net.res_bus_sc['ikss_ka'].at[sc_bus], 2)  # round(Switches['Fault Current'],2)

            line_text.append(text_line)

        # line annotations to collections for plotting
        line_annotate = create_annotation_collection(texts=line_text, coords=line_geodata, size=0.06, prop=None)
        collection.append(line_annotate)

        # Bus Annotations
        bus_text = []
        for i in net.bus.index:
            bus_texts = f'bus_{i}'
            bus_text.append(bus_texts)

        bus_text = bus_text[:-1]

        bus_geodata = net.bus.geo.apply(geojson.loads).apply(geojson.utils.coords).apply(next).to_list()

        # placing bus
        bus_geodata = [(x[0] - 0.11, x[1] + 0.095) for x in bus_geodata]

        bus_annotate = create_annotation_collection(texts=bus_text, coords=bus_geodata, size=0.06, prop=None)
        collection.append(bus_annotate)

        max_bus_idx = max(net.bus.dropna(subset=['geo']).index)

        # Short circuit annotations
        collection.append(get_fault_annotation(net, fault_current))

        # sc_location annotation
        collection.append(get_sc_location_annotation(net, sc_location))

        # switch annotations
        # from pandapower.protection.utility_functions import switch_geodata
        switch_text = []
        for switch_id in trip_decisions.switch_id:
            text_switch = r"sw_" + str(switch_id)
            switch_text.append(text_switch)

        switch_geodata = switch_geodatas(net, size=bus_size, distance_to_bus=3.25 * bus_size)
        for i, (x, y) in enumerate(switch_geodata):
            switch_geodata[i] = (x - 0.085, y + 0.055)
        switch_annotate = create_annotation_collection(
            texts=switch_text,
            coords=switch_geodata,
            size=0.06,
            prop=None
        )
        collection.append(switch_annotate)
    draw_collections(collection)


def calc_line_intersection(m1, b1, m2, b2):
    xi = (b1 - b2) / (m2 - m1)
    yi = m1 * xi + b1
    return xi, yi


# get connected lines using bus id
@deprecated("Use pandapower.get_connected_elements(net, 'line', bus_idx) instead!")
def get_connected_lines(net, bus_idx):
    return get_connected_elements(net, "line", bus_idx)


# Returns the index of the second bus an element is connected to, given a
# first one. E.g. the from_bus given the to_bus of a line.
@deprecated("Use pandapower.next_bus(net, bus, element_id instead!")
def next_buses(net, bus, element_id):
    return next_bus(net, bus, element_id)


# get the connected bus listr from start to end bus
def source_to_end_path(net, start_bus, bus_list, bus_order):
    connected_lines = get_connected_elements(net, 'line', start_bus)
    flag = 0
    for line in connected_lines:
        next_connected_bus = next_bus(net, bus=start_bus, element_id=line)
        bus_order_1 = bus_order.copy()
        if next_connected_bus in bus_order:
            continue
        else:
            bus_order.append(next_connected_bus)
            bus_list = source_to_end_path(net, next_connected_bus, bus_list, bus_order)

            bus_order = bus_order_1
            flag = 1
    if flag == 0:
        bus_list.append(bus_order)

    return bus_list


# get connected switches with bus
@deprecated("Use pandapower.get_connected_switches(net, buses, consider='l', status='closed') instead!")
def get_connected_switches(net, buses):
    return get_connected_switches(net, buses, consider='l', status="closed")


# get connected buses with a oven element
@deprecated(
    "Use pandapower.get_connected_buses_at_element(net, element, element_type='l', respect_in_service=False) instead!"
)
def connected_bus_in_line(net, element):
    return get_connected_buses_at_element(net, element, element_type='l', respect_in_service=False)


@deprecated("Use networkx topological search instead! See pandapower docs.")
def get_line_path(net, bus_path, sc_line_id=0):
    """line path from bus path """
    line_path = []
    for i in range(len(bus_path) - 1):
        bus1 = bus_path[i]
        bus2 = bus_path[i + 1]

        # line=net.line [(net.line.from_bus==bus1) & (net.line.to_bus==bus2)].index.item()

        line_path.extend(net.line[((net.line.from_bus == bus1) & (net.line.to_bus == bus2)) | (
                (net.line.from_bus == bus2) & (net.line.to_bus == bus1))].index.to_list())

    return line_path


def switch_geodatas(net, size, distance_to_bus):
    """get the coordinates for switches at middle of the line"""
    switch_geo = []
    switches = []

    if len(switches) == 0:
        lbs_switches = net.switch.index[net.switch.et == "l"]
    else:
        lbs_switches = switches

    for switch in lbs_switches:
        sb = net.switch.bus.loc[switch]
        line = net.line.loc[net.switch.element.loc[switch]]
        fb = line.from_bus
        tb = line.to_bus

        line_buses = {fb, tb}
        target_bus = list(line_buses - {sb})[0]

        pos_sb = _get_coords_from_bus_idx(net, sb)
        if len(pos_sb) > 1:
            ValueError(f'Bus {sb} has multiple geodata entries: {pos_sb}')
        if len(pos_sb) == 0:
            ValueError(f'Bus {sb} has no geodata entry.')
        pos_sb = pos_sb[0]
        pos_tb = np.zeros(2)

        pos_tb = _get_coords_from_bus_idx(net, target_bus)
        if len(pos_sb) > 1:
            ValueError(f'Bus {sb} has multiple geodata entries: {pos_sb}')
        if len(pos_sb) == 0:
            ValueError(f'Bus {sb} has no geodata entry.')
        pos_tb = pos_tb[0]

        # position of switch symbol
        vec = np.array(pos_tb) - np.array(pos_sb)
        mag = np.linalg.norm(vec)

        pos_sw = pos_sb + vec / mag * distance_to_bus
        switch_geo.append(pos_sw)
    return switch_geo


def create_I_t_plot(trip_decisions, switch_id):
    """function create I-T plot using tripping decisions"""
    if not MATPLOTLIB_INSTALLED:
        raise ImportError('matplotlib must be installed to run create_I_t_plot()')

    x = [0, 0, 0, 0]
    y = [0, 0, 0, 0]

    X = []
    for counter, switch_id in enumerate(switch_id):
        lst_I = [trip_decisions[switch_id]['Ig'], trip_decisions[switch_id]['Igg']]
        lst_t = [trip_decisions[switch_id]['tg'], trip_decisions[switch_id]['tgg']]
        fault_current = trip_decisions[switch_id]['Fault Current [kA]']

        label = 'Relay:R' + str(switch_id)
        x[counter] = [lst_I[0], lst_I[0], lst_I[1], lst_I[1], 100]
        y[counter] = [10, lst_t[0], lst_t[0], lst_t[1], lst_t[1]]

        X.append([x[counter], y[counter], label, fault_current])

    plt.figure()

    for count, data in enumerate(X):
        plt.loglog(X[count][0], X[count][1], label=X[count][2])

        plt.axvline(x=X[count][3], ymin=0, ymax=X[0][1][0], color='r')
        plt.text(X[count][3] + 0.01, 0.01, 'Ikss  ' + str(round(fault_current, 1)) + 'kA', rotation=0)

        plt.grid(True, which="both", ls="-")
        plt.title("I-t-plot")
        plt.xlabel("I [kA]")
        plt.ylabel("t [s]")
        plt.xlim(0.1, 11)
        plt.ylim(0.01, 11)

    plt.show()
    plt.legend()

    if MPLCURSORS_INSTALLED:
        # hover the plott
        cursor = mplcursors.cursor(hover=True)
        cursor.connect("add", lambda sel: sel.annotation.set_text(
            'I:{} kA,t:{} s'.format(round(sel.target[0], 2), round(sel.target[1], 2))))


def power_flow_end_points(net):
    """function calculate end point from meshed grid and the start point from the radial grid to ext grid"""

    pf_net = copy.deepcopy(net)
    runpp(pf_net)
    pf_loop_end_buses = []
    pf_radial_end_buses = []

    for bus in pf_net.bus.index:

        lines = get_connected_elements(pf_net, element_type='l', buses=bus)

        if len(lines) == 1:
            pf_radial_end_buses.append(bus)
        else:
            pf_endpoint = []
            for line in lines:
                if bus == pf_net.line.from_bus.at[line]:
                    if pf_net.res_line.p_from_mw[line] < 0:
                        endpoint = True
                    else:
                        endpoint = False
                else:
                    if pf_net.res_line.p_to_mw[line] < 0:
                        endpoint = True
                    else:
                        endpoint = False
                pf_endpoint.append(endpoint)

            if all(pf_endpoint):
                pf_loop_end_buses.append(bus)

    return pf_loop_end_buses, pf_radial_end_buses


def bus_path_from_to_bus(net, radial_start_bus, loop_start_bus, end_bus):
    """function calculate the bus path from start and end buses"""
    loop_path = []
    radial_path = []

    from pandapower.topology.create_graph import create_nxgraph
    G = create_nxgraph(net)

    for start_buses in radial_start_bus:
        radial = nx.shortest_path(G, source=start_buses, target=end_bus)
        radial_path.append(radial)

    for start_buses in loop_start_bus:
        for path in nx.all_simple_paths(G, source=start_buses, target=end_bus):
            loop_path.append(path)

    bus_path = radial_path + loop_path
    return bus_path


def get_switches_in_path(net, paths):
    """function calculate the switching times from the  bus path"""

    lines_in_path: List[List] = []

    for path in paths:
        lines_at_path: set = set()

        for bus in path:
            lines_at_path.update(get_connected_elements(net, "l", bus))

        lines_at_paths = [
            line for line in lines_at_path
            if net.line.from_bus[line] in path and net.line.to_bus[line] in path
        ]

        lines_in_path.append(lines_at_paths)

    switches_in_path = [
        [net.switch[(net.switch['et'] == 'l') & (net.switch['element'] == line)].index for line in line_path]
        for line_path in lines_in_path
    ]

    return switches_in_path


def get_vi_angle(net: pandapowerNet, switch_id: int, **kwargs) -> float:
    """calculate the angle between voltage and current with reference to voltage"""

    if "powerflow_results" in kwargs:
        logger.warning(
            "The powerflow_results argument is deprecated and will be removed in the future."
        )

    runpp(net)
    line_idx = net.switch.element.at[switch_id]

    if get_from_bus_info_switch(net, switch_id):
        p = net.res_line_sc.p_from_mw.at[line_idx]
        q = net.res_line_sc.q_from_mvar.at[line_idx]
    else:
        p = net.res_line_sc.p_to_mw.at[line_idx]
        q = net.res_line_sc.q_to_mvar.at[line_idx]

    if p > 0 and q > 0:
        vi_angle = math.degrees(math.atan(q / p))
    elif p < 0 <= q:
        vi_angle = math.degrees(math.atan(q / p)) + 180
    elif p < 0 and q < 0:
        vi_angle = math.degrees(math.atan(q / p)) - 180
    elif p == 0 < q:
        vi_angle = 90
    elif p == 0 > q:
        vi_angle = -90
    else:
        vi_angle = math.inf
    return vi_angle


def bus_path_multiple_ext_bus(net):
    G = create_nxgraph(net)
    bus_path = []
    for line_id in net.line.index:

        # line_id=62
        from_bus = net.line.from_bus.at[line_id]
        to_bus = net.line.to_bus.at[line_id]
        max_bus_path = []

        if net.trafo.empty:
            # for ext_bus in net.ext_grid.bus:
            for ext_bus in set(net.ext_grid.bus):
                from_bus_path = nx.shortest_path(G, source=ext_bus, target=from_bus)
                to_bus_path = nx.shortest_path(G, source=ext_bus, target=to_bus)

                if len(from_bus_path) == len(to_bus_path):
                    from_bus_path.append(to_bus_path[-1])
                    max_bus_path.append(from_bus_path)

                elif len(from_bus_path) != len(to_bus_path):
                    if len(from_bus_path) > 1 and len(to_bus_path) > 1:
                        min_len = min(len(from_bus_path), len(to_bus_path))
                        if from_bus_path[min_len - 1] != to_bus_path[min_len - 1]:
                            if len(from_bus_path) < len(to_bus_path):
                                from_bus_path.append(to_bus_path[-1])
                                max_bus_path.append(from_bus_path)
                            else:
                                to_bus_path.append(from_bus_path[-1])
                                max_bus_path.append(to_bus_path)
                        else:
                            max_bus_path.append(max([from_bus_path, to_bus_path]))
                    else:
                        max_bus_path.append(max([from_bus_path, to_bus_path]))

            bus_path.append(sorted(max_bus_path, key=len)[0])

    return bus_path


# get the line path from the given bus path
def get_line_path(net, bus_path):
    """ Function return the list of line path from the given bus path"""
    line_path = []
    for i in range(len(bus_path) - 1):
        bus1 = bus_path[i]
        bus2 = bus_path[i + 1]
        line1 = net.line[(net.line.from_bus == bus1) & (net.line.to_bus == bus2)].index.to_list()
        line2 = net.line[(net.line.from_bus == bus2) & (net.line.to_bus == bus1)].index.to_list()
        if len(line2) == 0:
            line_path.append(line1[0])

        if len(line1) == 0:
            line_path.append(line2[0])
    return line_path


def parallel_lines(net):
    """ Function return the list of parallel lines in the network"""

    parallel = []
    new_parallel = []

    # parallel_lines
    for i in net.line.index:
        for j in net.line.index:
            if i == j:
                continue

            i_from_bus = net.line.loc[i].from_bus
            i_to_bus = net.line.loc[i].to_bus
            j_from_bus = net.line.loc[j].from_bus
            j_to_bus = net.line.loc[j].to_bus

            if (
                    (i_from_bus == j_from_bus and i_to_bus == j_to_bus)
                    or (i_from_bus == j_to_bus and i_to_bus == j_from_bus)
                    or (i_to_bus == j_from_bus and i_from_bus == j_to_bus)
                    or (i_to_bus == j_to_bus and i_from_bus == j_from_bus)
            ):
                parallel.append([i, j])

        parallel_line = [list(i) for i in set(map(tuple, parallel))]

        # remove duplicates
        new_parallel = []
        for line in parallel_line:
            if line not in new_parallel:
                new_parallel.append(line)

    return new_parallel


def read_fuse_from_std_type():
    # write characteristic of fuse from fuse_std_library
    # net,overwrite already existing characteristic
    # raise error if name of fuse is not available in fuse library
    raise NotImplementedError("This function is not implemented yet.")



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
from typing import Union, Dict, List
import pandas as pd
import logging
import time
from . import cim_tools

from pandapower.auxiliary import pandapowerNet
from pandapower.std_types import create_std_type
from pandapower.create import create_bus, create_ext_grid, create_lines

logger = logging.getLogger(__name__)


def _set_column_to_type(input_df: pd.DataFrame, column: str, data_type):
    try:
        input_df[column] = input_df[column].astype(data_type)
    except Exception as e:
        logger.error("Couldn't set data type %s for column %s!" % (data_type, column))
        logger.exception(e)


def set_pp_col_types(net: Union[pandapowerNet, Dict], ignore_errors: bool = False) -> pandapowerNet:
    """
    Set the data types for some columns from pandapower assets. This mainly effects bus columns (to int, e.g.
    sgen.bus or line.from_bus) and in_service and other boolean columns (to bool, e.g. line.in_service or gen.slack).
    :param net: The pandapower network to update the data types.
    :param ignore_errors: Ignore problems if set to True (no warnings displayed). Optional, default: False.
    :return: The pandapower network with updated data types.
    """
    time_start = time.time()
    pp_elements = ['bus', 'dcline', 'ext_grid', 'gen', 'impedance', 'line', 'load', 'motor', 'sgen', 'shunt', 'storage',
                   'switch', 'trafo', 'trafo3w', 'ward', 'xward']
    to_int = ['bus', 'element', 'to_bus', 'from_bus', 'hv_bus', 'mv_bus', 'lv_bus']
    to_bool = ['in_service', 'closed']
    logger.info("Setting the columns data types for buses to int and in_service to bool for the following elements: "
                "%s" % pp_elements)
    int_type = int
    bool_type = bool
    for ele in pp_elements:
        logger.info("Accessing pandapower element %s." % ele)
        if not hasattr(net, ele):
            if not ignore_errors:
                logger.warning("Missing the pandapower element %s in the input pandapower network!" % ele)
            continue
        for one_int in to_int:
            if one_int in net[ele].columns:
                _set_column_to_type(net[ele], one_int, int_type)
        for one_bool in to_bool:
            if one_bool in net[ele].columns:
                _set_column_to_type(net[ele], one_bool, bool_type)
    # some individual things
    if hasattr(net, 'sgen'):
        _set_column_to_type(net['sgen'], 'current_source', bool_type)
    if hasattr(net, 'gen'):
        _set_column_to_type(net['gen'], 'slack', bool_type)
    if hasattr(net, 'shunt'):
        _set_column_to_type(net['shunt'], 'step', int_type)
        _set_column_to_type(net['shunt'], 'max_step', int_type)
    logger.info("Finished setting the data types for the pandapower network in %ss." % (time.time() - time_start))
    return net


def add_slack_and_lines_to_boundary_nodes(net: pandapowerNet, voltage_levels: List[int] = None):
    """
    Add lines with low impedance and a slack to the boundary nodes with the highest voltage.
    :param net: The pandapower network
    :param voltage_levels: The voltage levels to add lines and slacks. For each given voltage level, lines and one
    slack will be connected to the corresponding boundary nodes. Optional, default: Highest voltage level from
    boundary nodes.
    :return:
    """
    sc = cim_tools.get_pp_net_special_columns_dict()
    busses = net.bus.loc[(net.bus[sc['o_prf']] == 'eq_bd') | (net.bus[sc['o_prf']] == 'tp_bd')]
    if voltage_levels is None:
        max_voltage = busses['vn_kv'].max()
        logger.info("Highest voltage level: %skV" % max_voltage)
        voltage_levels = [max_voltage]
    create_std_type(net, data=dict({'r_ohm_per_km': 0, 'x_ohm_per_km': .05, 'c_nf_per_km': 0, 'max_i_ka': 9999}),
                       name='low_impedance_line', element='line')
    for one_voltage_level in voltage_levels:
        logger.info("Processing voltage level %skV" % one_voltage_level)
        busses_t = busses.loc[busses['vn_kv'] == one_voltage_level]
        new_bus_id = create_bus(net, vn_kv=one_voltage_level,
                                   name='virtual slack bus at voltage level ' + str(one_voltage_level))
        logger.info("Added virtual slack bus with ID: %s" % new_bus_id)
        create_ext_grid(net, bus=new_bus_id, vm_pu=1.0,
                           name='virtual slack at voltage level ' + str(one_voltage_level))
        logger.info("Added slack at bus ID: %s" % new_bus_id)
        new_bus_id_array = [new_bus_id for _ in busses_t.index.values]
        create_lines(net, from_buses=busses_t.index.values, to_buses=new_bus_id_array, std_type='low_impedance_line',
                        name='virtual line to slack node with voltage level ' + str(one_voltage_level), length_km=1)
        logger.info("Created %s low impedance lines." % len(busses_t.index.values))
        del busses_t


def get_not_existing_column(df: pd.DataFrame) -> str:
    """
    Get a not existing column name in a DataFrame
    :param df: The DataFrame where to get a not existing column
    :return: The name of a not existing column, e.g. temp_col_0 or temp_col_1
    """
    i = 0
    col = 'temp_col_'
    while col + str(i) in df.columns:
        i += 1
    return col + str(i)



# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""Builds the vector of complex bus power injections.
"""

from numpy import ones, flatnonzero as find
from pandapower.pypower.idx_bus import PD, QD
from pandapower.pypower.idx_gen import GEN_BUS, PG, QG, GEN_STATUS
from scipy.sparse import csr_matrix as sparse

from pandapower.pypower.idx_bus import CID, CZD


def _get_Sbus(baseMVA, bus, gen_on, Cg, vm=None):
    # power injected by gens plus power injected by loads converted to p.u.
    S_load = _get_Sload(bus, vm)
    Sbus = (Cg * (gen_on[:, PG] + 1j * gen_on[:, QG])
            - S_load) / baseMVA
    return Sbus


def _get_Sload(bus, vm):
    S_load = bus[:, PD] + 1j * bus[:, QD]
    if vm is not None:
        ci = bus[:, CID]
        cz = bus[:, CZD]
        cp = (1 - ci - cz)
        volt_depend = cp + ci * vm + cz * vm ** 2
        S_load *= volt_depend
    return S_load


def _get_Cg(gen_on, bus):
    gbus = gen_on[:, GEN_BUS]  ## what buses are they at?

    ## form net complex bus power injection vector
    nb = bus.shape[0]
    ngon = gen_on.shape[0]
    return sparse((ones(ngon), (gbus, range(ngon))), (nb, ngon))


def makeSbus(baseMVA, bus, gen, vm=None):
    """Builds the vector of complex bus power injections.

    Returns the vector of complex bus power injections, that is, generation
    minus load. Power is expressed in per unit.

    @see: L{makeYbus}

    @author: Ray Zimmerman (PSERC Cornell)
    @author: Richard Lincoln
    """
    ## generator info
    on = find(gen[:, GEN_STATUS] > 0)  ## which generators are on?
    gen_on = gen[on, :]

    ## connection matrix, element i, j is 1 if gen on(j) at bus i is ON
    Cg = _get_Cg(gen_on, bus)

    ## power injected by gens plus power injected by loads converted to p.u.
    Sbus = _get_Sbus(baseMVA, bus, gen_on, Cg, vm)

    return Sbus



from io import StringIO
import networkx as nx
import numpy as np
import pandas as pd
from pandas import testing as pdt
from deepdiff import DeepDiff

from pandapower.auxiliary import pandapowerNet

try:
    from networkx.utils.misc import graphs_equal
    GRAPHS_EQUAL_POSSIBLE = True
except ImportError:
    GRAPHS_EQUAL_POSSIBLE = False

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def dataframes_equal(df1, df2, ignore_index_order=True, **kwargs):
    """
    Returns a boolean whether the given two dataframes are equal or not.
    """
    if "tol" in kwargs:
        if "atol" in kwargs:
            raise ValueError("'atol' and 'tol' are given to dataframes_equal(). Don't use 'tol' "
                             "anymore.")
        logger.warning("in dataframes_equal() parameter 'tol' is deprecated. Use 'atol' instead.")
        kwargs["atol"] = kwargs.pop("tol")

    if ignore_index_order:
        df1 = df1.sort_index().sort_index(axis=1)
        df2 = df2.sort_index().sort_index(axis=1)

    # geo columns are compared later
    df1_cols = df1.columns.difference({"geo"})
    df2_cols = df2.columns.difference({"geo"})

    # --- pandas implementation
    try:
        pdt.assert_frame_equal(df1[df1_cols], df2[df2_cols], **kwargs)
    except AssertionError:
        return False

    # --- compare geo columns
    if "geo" in df1.columns and "geo" in df2.columns:
        not_eq_warn = "DataFrames do not match in column 'geo'."
        notnull1 = df1.geo.index[~df1.geo.isnull()]
        notnull1 = df1.geo.loc[notnull1].index[df1.geo.loc[notnull1].apply(len).astype(bool)]
        notnull2 = df2.geo.index[~df2.geo.isnull()]
        notnull2 = df2.geo.loc[notnull2].index[df2.geo.loc[notnull2].apply(len).astype(bool)]

        if len(notnull1) + len(notnull2) == 0:
            return True

        # check equal index with geo data
        if len(notnull1.symmetric_difference(notnull2)):
            logger.warning(not_eq_warn)
            return False

        # create DataFrames from geo information
        df1_geo = pd.concat([pd.read_json(StringIO(df1.geo.at[idx])).assign(**{"idx": idx}) for idx
                             in notnull1])
        df2_geo = pd.concat([pd.read_json(StringIO(df2.geo.at[idx])).assign(**{"idx": idx}) for idx
                             in notnull2])

        # check equal columns and reorder columns of df2_geo
        if len(df1_geo.columns.symmetric_difference(df2_geo.columns)):
            logger.warning(not_eq_warn)
            return False
        else:
            df2_geo = df2_geo[df1_geo.columns]

        try:
            pdt.assert_frame_equal(df1_geo, df2_geo, **kwargs)
            return True
        except AssertionError:
            logger.warning(not_eq_warn)
            return False
        return geo_eq

    elif "geo" not in df1.columns and "geo" not in df2.columns:
        return True
    elif "geo" not in df1.columns and "geo" in df2.columns:
        logger.warning("df2 contains column 'geo' but df1 doesn't.")
        return False
    else:
        logger.warning("df1 contains column 'geo' but df2 doesn't.")
        return False


def compare_arrays(x, y):
    """
    Returns an array of bools whether array x is equal to array y. Strings are allowed in x
    or y. NaN values are assumed as equal.
    """
    if x.shape == y.shape:
        # (x != x) is like np.isnan(x) - but works also for strings
        return np.equal(x, y) | ((x != x) & (y != y))
    else:
        raise ValueError("x and y need to have the same shape.")


def nets_equal(net1, net2, check_only_results=False, check_without_results=False, exclude_elms=None,
               name_selection=None, **kwargs):
    """
    Returns a boolean whether the two given pandapower networks are equal.

    pandapower net keys starting with "_" are ignored. Same for the key "et" (elapsed time).

    If the element tables contain JSONSerializableClass objects, they will also be compared:
    attributes are compared but not the addresses of the objects.

    INPUT:
        **net1** (pandapowerNet)

        **net2** (pandapowerNet)

    OPTIONAL:
        **check_only_results** (bool, False) - if True, only result tables (starting with ``res_``)
        are compared

        **check_without_results** (bool, False) - if True, result tables (starting with ``res_``)
        are ignored for comparison

        **exclude_elms** (list, None) - list of element tables which should be ignored in the
        comparison

        **name_selection** (list, None) - list of element tables which should be compared

        **kwargs** - key word arguments for dataframes_equal()
    """
    if not (isinstance(net1, pandapowerNet) and isinstance(net2, pandapowerNet)):
        logger.warning("At least one net is not of type pandapowerNet.")
        return False
    not_equal, not_checked_keys = nets_equal_keys(
        net1, net2, check_only_results, check_without_results, exclude_elms, name_selection,
        **kwargs)
    if len(not_checked_keys) > 0:
        logger.warning("These keys were ignored by the comparison of the networks: %s" % (', '.join(
            not_checked_keys)))

    if len(not_equal) > 0:
        logger.warning("Networks do not match in DataFrame(s): %s" % (', '.join(not_equal)))
        return False
    else:
        return True


def nets_equal_keys(net1, net2, check_only_results, check_without_results, exclude_elms,
                     name_selection, **kwargs):
    """ Returns a lists of keys which are 1) not equal and 2) not checked.
    Used within nets_equal(). """
    if check_without_results and check_only_results:
        raise UserWarning("Please provide only one of the options to check without results or to "
                          "exclude results in comparison.")

    exclude_elms = [] if exclude_elms is None else list(exclude_elms)
    exclude_elms += ["res_" + ex for ex in exclude_elms]
    not_equal = []

    # for two networks make sure both have the same keys
    if name_selection is not None:
        net1_keys = net2_keys = name_selection
    elif check_only_results:
        net1_keys = [key for key in net1.keys() if key.startswith("res_")
                     and key not in exclude_elms]
        net2_keys = [key for key in net2.keys() if key.startswith("res_")
                     and key not in exclude_elms]
    else:
        net1_keys = [key for key in net1.keys() if not (
            key.startswith("_") or key in exclude_elms or key == "et"
            or key.startswith("res_") and check_without_results)]
        net2_keys = [key for key in net2.keys() if not (
            key.startswith("_") or key in exclude_elms or key == "et"
            or key.startswith("res_") and check_without_results)]
    keys_to_check = set(net1_keys) & set(net2_keys)
    key_difference = set(net1_keys) ^ set(net2_keys)
    not_checked_keys = list()

    if len(key_difference) > 0:
        logger.warning(f"Networks entries mismatch at: {key_difference}")
        return key_difference, set()

    # ... and then iter through the keys, checking for equality for each table
    for key in list(keys_to_check):

        if isinstance(net1[key], pd.DataFrame):
            if not isinstance(net2[key], pd.DataFrame) or not dataframes_equal(
                    net1[key], net2[key], **kwargs):
                not_equal.append(key)

        elif isinstance(net1[key], np.ndarray):
            if not isinstance(net2[key], np.ndarray):
                not_equal.append(key)
            else:
                if not np.array_equal(net1[key], net2[key], equal_nan=True):
                    not_equal.append(key)

        elif isinstance(net1[key], int) or isinstance(net1[key], float) or \
                isinstance(net1[key], complex):
            if not np.isclose(net1[key], net2[key]):
                not_equal.append(key)

        elif isinstance(net1[key], nx.Graph):
            if GRAPHS_EQUAL_POSSIBLE:
                if not graphs_equal(net1[key], net2[key]):
                    not_equal.append(key)
            else:
                # Maybe there is a better way, but at least this could be checked
                if net1[key].nodes != net2[key].nodes or net1[key].edges != net2[key].edges:
                    not_equal.append(key)
        elif isinstance(net1[key], dict):
            diff = DeepDiff(net1[key], net2[key], math_epsilon=1e-20, ignore_numeric_type_changes=True)
            if len(diff) > 0:
                not_equal.append(key)

        else:
            try:
                is_eq = net1[key] == net2[key]
                if not is_eq:
                    not_equal.append(key)
            except:
                not_checked_keys.append(key)
    return not_equal, not_checked_keys



# -*- coding: utf-8 -*-nt

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from copy import deepcopy
import os
import json
from functools import reduce
from typing import Optional, Union
import numpy as np
import pandas as pd
from pandas.api.types import is_integer_dtype, is_object_dtype
from pandapower.io_utils import pandapowerNet
from pandapower.create import create_empty_network, create_buses, create_lines_from_parameters, \
    create_transformers_from_parameters
from pandapower.topology import create_nxgraph, connected_components
from pandapower.plotting import set_line_geodata_from_bus_geodata
from pandapower.toolbox import drop_buses, fuse_buses

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def from_jao(excel_file_path: str,
             html_file_path: Optional[str],
             extend_data_for_grid_group_connections: bool,
             drop_grid_groups_islands: bool = False,
             apply_data_correction: bool = True,
             max_i_ka_fillna: Union[float, int] = 999,
             **kwargs) -> pandapowerNet:
    """Converts European (Core) EHV grid data provided by JAO (Joint Allocation Office), the
    "Single Allocation Platform (SAP) for all European Transmission System Operators (TSOs) that
    operate in accordance to EU legislation".

    **Data Sources and Availability:**

    The data are available at the website
    `JAO Static Grid Model <https://www.jao.eu/static-grid-model>`_ (November 2024).
    There, a map is provided to get an fine overview of the geographical extent and the scope of
    the data. These inlcude information about European (Core) lines, tielines, and transformers.

    **Limitations:**

    No information is available on load or generation.
    The data quality with regard to the interconnection of the equipment, the information provided
    and the (incomplete) geodata should be considered with caution.

    **Features of the converter:**

    - **Data Correction:** corrects known data inconsistencies, such as inconsistent spellings and missing necessary information.
    - **Geographical Data Parsing:** Parses geographical data from the HTML file to add geolocation information to buses and lines.
    - **Grid Group Connections:** Optionally extends the network by connecting islanded grid groups to avoid disconnected components.
    - **Data Customization:** Allows for customization through additional parameters to control transformer creation, grid group dropping, and voltage level deviations.

    Parameters
    ----------
    excel_file_path : str
        input data including electrical parameters of grids' utilities, stored in multiple sheets
        of an excel file
    html_file_path : str
        input data for geo information. If The converter should be run without geo information, None
        can be passed., provided by an html file
    extend_data_for_grid_group_connections : bool
        if True, connections (additional transformers and merging buses) are created to avoid
        islanded grid groups, by default False
    drop_grid_groups_islands : bool, optional
        if True, islanded grid groups will be dropped if their number of buses is below
        min_bus_number (default is 6), by default False
    apply_data_correction : bool, optional
        _description_, by default True
    max_i_ka_fillna : float | int, optional
        value to fill missing values or data of false type in max_i_ka of lines and transformers.
        If no value should be set, you can also pass np.nan. By default 999

    Returns
    -------
    pandapowerNet
        net created from the jao data

    Additional Parameters
    ---------------------
    minimal_trafo_invention : bool, optional
        applies if extend_data_for_grid_group_connections is True. Then, if minimal_trafo_invention
        is True, adding transformers stops when no grid groups is islanded anymore (does not apply
        for release version 5 or 6, i.e. it does not care what value is passed to
        minimal_trafo_invention). If False, all equally named buses that have different voltage
        level and lay in different groups will be connected via additional transformers,
        by default False
    min_bus_number : Union[int,str], optional
        Threshold value to decide which small grid groups should be dropped and which large grid
        groups should be kept. If all islanded grid groups should be dropped except of the one
        largest, set "max". If all grid groups that do not contain a slack element should be
        dropped, set "unsupplied". By default 6
    rel_deviation_threshold_for_trafo_bus_creation : float, optional
        If the voltage level of transformer locations is far different than the transformer data,
        additional buses are created. rel_deviation_threshold_for_trafo_bus_creation defines the
        tolerance in which no additional buses are created. By default 0.2
    log_rel_vn_deviation : float, optional
        This parameter allows a range below rel_deviation_threshold_for_trafo_bus_creation in which
        a warning is logged instead of a creating additional buses. By default 0.12

    Examples
    --------
    >>> from pathlib import Path
    >>> import os
    >>> import pandapower as pp
    >>> net = pp.converter.from_jao()
    >>> home = str(Path.home())
    >>> # assume that the files are located at your desktop:
    >>> excel_file_path = os.path.join(home, "desktop", "202409_Core Static Grid Mode_6th release")
    >>> html_file_path = os.path.join(home, "desktop", "2024-09-13_Core_SGM_publication.html")
    >>> net = from_jao(excel_file_path, html_file_path, True, drop_grid_groups_islands=True)
    """

    # --- read data
    data = pd.read_excel(excel_file_path, sheet_name=None, header=[0, 1])
    if html_file_path is not None:
        with open(html_file_path, mode='r', encoding=kwargs.get("encoding", "utf-8")) as f:
            html_str = f.read()
    else:
        html_str = ""

    # --- manipulate data / data corrections
    if apply_data_correction:
        html_str = _data_correction(data, html_str, max_i_ka_fillna)

    # --- parse html_str to line_geo_data
    line_geo_data = None
    if html_str:
        try:
            line_geo_data = _parse_html_str(html_str)
        except (json.JSONDecodeError, KeyError, AssertionError) as e:
            logger.error(f"html data were ignored due to this error:\n{e}")

    # --- create the pandapower net
    net = create_empty_network(name=os.path.splitext(os.path.basename(excel_file_path))[0],
                               **{key: val for key, val in kwargs.items() if key == "sn_mva"})
    _create_buses_from_line_data(net, data)
    _create_lines(net, data, max_i_ka_fillna)
    _create_transformers_and_buses(net, data, **kwargs)

    # --- invent connections between grid groups
    if extend_data_for_grid_group_connections:
        _invent_connections_between_grid_groups(net, **kwargs)

    # --- drop islanded grid groups
    if drop_grid_groups_islands:
        drop_islanded_grid_groups(net, kwargs.get("min_bus_number", 6))

    # --- add geo data to buses and lines
    if line_geo_data is not None:
        _add_bus_geo(net, line_geo_data)
        set_line_geodata_from_bus_geodata(net)

    return net

# --- secondary functions --------------------------------------------------------------------------


def _data_correction(
        data: dict[str, pd.DataFrame],
        html_str: Optional[str],
        max_i_ka_fillna: Union[float, int]) -> Optional[str]:
    """Corrects input data in particular with regard to obvious weaknesses in the data provided,
    such as inconsistent spellings and missing necessary information

    Parameters
    ----------
    data : dict[str, pd.DataFrame]
        data provided by the excel file which will be corrected
    html_str : str | None
        data provided by the html file which will be corrected
    max_i_ka_fillna : float | int
        value to fill missing values or data of false type in max_i_ka of lines and transformers.
        If no value should be set, you can also pass np.nan.

    Returns
    -------
    str
        corrected html_str
    """
    # old name -> new name
    rename_locnames = [("PSTMIKULOWA", "PST MIKULOWA"),
                       ("Chelm", "CHELM"),
                       ("OLSZTYN-MATK", "OLSZTYN-MATKI"),
                       ("STANISLAWOW", "Stanislawow"),
                       ("VIERRADEN", "Vierraden")]

    # --- Line and Tieline data ---------------------------
    for key in ["Lines", "Tielines"]:

        # --- correct column names
        cols = data[key].columns.to_frame().reset_index(drop=True)
        cols.loc[cols[1] == "Voltage_level(kV)", 0] = None
        cols.loc[cols[1] == "Comment", 0] = None
        cols.loc[cols[0].str.startswith("Unnamed:").astype(bool), 0] = None
        cols.loc[cols[1] == "Length_(km)", 0] = "Electrical Parameters"  # might be wrong in
        # Tielines otherwise
        data[key].columns = pd.MultiIndex.from_arrays(cols.values.T)

        # --- correct comma separation and cast to floats
        data[key][("Maximum Current Imax (A)", "Fixed")] = \
            data[key][("Maximum Current Imax (A)", "Fixed")].replace(
                "\xa0", max_i_ka_fillna*1e3).replace(
                "-", max_i_ka_fillna*1e3).replace(" ", max_i_ka_fillna*1e3)
        col_names = [("Electrical Parameters", col_level1) for col_level1 in [
            "Length_(km)", "Resistance_R(Ω)", "Reactance_X(Ω)", "Susceptance_B(μS)",
            "Length_(km)"]] + [("Maximum Current Imax (A)", "Fixed")]
        _float_col_comma_correction(data, key, col_names)

        # --- consolidate to one way of name capitalization
        for loc_name in [(None, "NE_name"), ("Substation_1", "Full_name"),
                         ("Substation_2", "Full_name")]:
            data[key].loc[:, loc_name] = data[key].loc[:, loc_name].str.strip().apply(
                _multi_str_repl, repl=rename_locnames)
    html_str = _multi_str_repl(html_str, rename_locnames)

    # --- Transformer data --------------------------------
    key = "Transformers"

    # --- fix Locations
    loc_name = ("Location", "Full Name")
    data[key].loc[:, loc_name] = data[key].loc[:, loc_name].str.strip().apply(
        _multi_str_repl, repl=rename_locnames)

    # --- fix data in nonnull_taps
    taps = data[key].loc[:, ("Phase Shifting Properties", "Taps used for RAO")].fillna("").astype(
        str).str.replace(" ", "")
    nonnull = taps.apply(len).astype(bool)
    nonnull_taps = taps.loc[nonnull]
    surrounded = nonnull_taps.str.startswith("<") & nonnull_taps.str.endswith(">")
    nonnull_taps.loc[surrounded] = nonnull_taps.loc[surrounded].str[1:-1]
    slash_sep = (~nonnull_taps.str.contains(";")) & nonnull_taps.str.contains("/")
    nonnull_taps.loc[slash_sep] = nonnull_taps.loc[slash_sep].str.replace("/", ";")
    nonnull_taps.loc[nonnull_taps == "0"] = "0;0"
    data[key].loc[nonnull, ("Phase Shifting Properties", "Taps used for RAO")] = nonnull_taps
    data[key].loc[~nonnull, ("Phase Shifting Properties", "Taps used for RAO")] = "0;0"

    # --- phase shifter with double info
    cols = ["Phase Regulation δu (%)", "Angle Regulation δu (%)"]
    for col in cols:
        if is_object_dtype(data[key].loc[:, ("Phase Shifting Properties", col)]):
            tr_double = data[key].index[data[key].loc[:, (
                "Phase Shifting Properties", col)].str.contains("/").fillna(0).astype(bool)]
            data[key].loc[tr_double, ("Phase Shifting Properties", col)] = data[key].loc[
                tr_double, ("Phase Shifting Properties", col)].str.split("/", expand=True)[
                1].str.replace(",", ".").astype(float).values  # take second info and correct
                # separation: , -> .

    return html_str


def _parse_html_str(html_str: str) -> pd.DataFrame:
    """Converts ths geodata from the html file (information hidden in the string), from Lines in
    particular, to a DataFrame that can be used later in _add_bus_geo()

    Parameters
    ----------
    html_str : str
        html file that includes geodata information

    Returns
    -------
    pd.DataFrame
        extracted geodata for a later and easy use
    """
    def _filter_name(st: str) -> str:
        name_start = "<b>NE name: "
        name_end = "</b>"
        pos0 = st.find(name_start) + len(name_start)
        pos1 = st.find(name_end, pos0)
        assert pos0 >= 0
        assert pos1 >= len(name_start)
        return st[pos0:pos1]

    json_start_str = '<script type="application/json" data-for="htmlwidget-216030e6806f328c00fb">'
    json_start_pos = html_str.find(json_start_str) + len(json_start_str)
    json_end_pos = html_str[json_start_pos:].find('</script>')
    json_str = html_str[json_start_pos:(json_start_pos+json_end_pos)]
    geo_data = json.loads(json_str)
    geo_data = geo_data["x"]["calls"]
    methods_pos = pd.Series({item["method"]: i for i, item in enumerate(geo_data)})
    polylines = geo_data[methods_pos.at["addPolylines"]]["args"]
    EIC_start = "EIC Code:<b> "
    if len(polylines[6]) != len(polylines[0]):
        raise AssertionError("The lists of EIC Code data and geo data are not of the same length.")
    line_EIC = [polylines[6][i][polylines[6][i].find(EIC_start)+len(EIC_start):] for i in range(
        len(polylines[6]))]
    line_name = [_filter_name(polylines[6][i]) for i in range(len(polylines[6]))]
    line_geo_data = pd.concat([_lng_lat_to_df(polylines[0][i][0][0], line_EIC[i], line_name[i]) for
                               i in range(len(polylines[0]))], ignore_index=True)

    # remove trailing whitespaces
    for col in ["EIC_Code", "name"]:
        line_geo_data[col] = line_geo_data[col].str.strip()

    return line_geo_data


def _create_buses_from_line_data(net: pandapowerNet, data: dict[str, pd.DataFrame]) -> None:
    """Creates buses to the pandapower net using information from the lines and tielines sheets
    (excel file).

    Parameters
    ----------
    net : pandapowerNet
        net to be filled by buses
    data : dict[str, pd.DataFrame]
        data provided by the excel file which will be corrected
    """
    bus_df_empty = pd.DataFrame({"name": str(), "vn_kv": float(), "TSO": str()}, index=[])
    bus_df = deepcopy(bus_df_empty)
    for key in ["Lines", "Tielines"]:
        for subst in ['Substation_1', 'Substation_2']:
            data_col_tuples = [(subst, "Full_name"), (None, "Voltage_level(kV)"), (None, "TSO")]
            to_add = data[key].loc[:, data_col_tuples].set_axis(bus_df.columns, axis="columns")
            if len(bus_df):
                bus_df = pd.concat([bus_df, to_add])
            else:
                bus_df = to_add
    bus_df = _drop_duplicates_and_join_TSO(bus_df)
    new_bus_idx = create_buses(
        net, len(bus_df), vn_kv=bus_df.vn_kv, name=bus_df.name, zone=bus_df.TSO)
    assert np.allclose(new_bus_idx, bus_df.index)


def _create_lines(
        net: pandapowerNet,
        data: dict[str, pd.DataFrame],
        max_i_ka_fillna: Union[float, int]) -> None:
    """Creates lines to the pandapower net using information from the lines and tielines sheets
    (excel file).

    Parameters
    ----------
    net : pandapowerNet
        net to be filled by buses
    data : dict[str, pd.DataFrame]
        data provided by the excel file which will be corrected
    max_i_ka_fillna : float | int
        value to fill missing values or data of false type in max_i_ka of lines and transformers.
        If no value should be set, you can also pass np.nan.
    """

    bus_idx = _get_bus_idx(net)

    for key in ["Lines", "Tielines"]:
        length_km = data[key][("Electrical Parameters", "Length_(km)")].values
        zero_length = np.isclose(length_km, 0)
        no_length = np.isnan(length_km)
        if sum(zero_length) or sum(no_length):
            logger.warning(f"According to given data, {sum(zero_length)} {key.lower()} have zero "
                           f"length and {sum(zero_length)} {key.lower()} have no length data. "
                           "Both types of wrong data are replaced by 1 km.")
            length_km[zero_length | no_length] = 1
        vn_kvs = data[key].loc[:, (None, "Voltage_level(kV)")].values

        _ = create_lines_from_parameters(
            net,
            bus_idx.loc[list(tuple(zip(data[key].loc[:, ("Substation_1", "Full_name")].values,
                        vn_kvs)))].values,
            bus_idx.loc[list(tuple(zip(data[key].loc[:, ("Substation_2", "Full_name")].values,
                        vn_kvs)))].values,
            length_km,
            data[key][("Electrical Parameters", "Resistance_R(Ω)")].values / length_km,
            data[key][("Electrical Parameters", "Reactance_X(Ω)")].values / length_km,
            data[key][("Electrical Parameters", "Susceptance_B(μS)")].values / length_km,
            data[key][("Maximum Current Imax (A)", "Fixed")].fillna(
                max_i_ka_fillna*1e3).values / 1e3,
            name=data[key].xs("NE_name", level=1, axis=1).values[:, 0],
            EIC_Code=data[key].xs("EIC_Code", level=1, axis=1).values[:, 0],
            TSO=data[key].xs("TSO", level=1, axis=1).values[:, 0],
            Comment=data[key].xs("Comment", level=1, axis=1).values[:, 0],
            Tieline=key == "Tielines",
        )


def _create_transformers_and_buses(
        net: pandapowerNet, data: dict[str, pd.DataFrame], **kwargs) -> None:
    """Creates transformers to the pandapower net using information from the transformers sheet
    (excel file).

    Parameters
    ----------
    net : pandapowerNet
        net to be filled by buses
    data : dict[str, pd.DataFrame]
        data provided by the excel file which will be corrected
    """

    # --- data preparations
    key = "Transformers"
    bus_idx = _get_bus_idx(net)
    vn_hv_kv, vn_lv_kv = _get_transformer_voltages(data, bus_idx)
    trafo_connections = _allocate_trafos_to_buses_and_create_buses(
        net, data, bus_idx, vn_hv_kv, vn_lv_kv, **kwargs)
    max_i_a = data[key].loc[:, ("Maximum Current Imax (A) primary", "Fixed")]
    empty_i_idx = max_i_a.index[max_i_a.isnull()]
    max_i_a.loc[empty_i_idx] = data[key].loc[empty_i_idx, (
        "Maximum Current Imax (A) primary", "Max")].values
    sn_mva = np.sqrt(3) * max_i_a * vn_hv_kv / 1e3
    z_pu = vn_lv_kv**2 / sn_mva
    rk = data[key].xs("Resistance_R(Ω)", level=1, axis=1).values[:, 0] / z_pu
    xk = data[key].xs("Reactance_X(Ω)", level=1, axis=1).values[:, 0] / z_pu
    b0 = data[key].xs("Susceptance_B (µS)", level=1, axis=1).values[:, 0] * 1e-6 * z_pu
    g0 = data[key].xs("Conductance_G (µS)", level=1, axis=1).values[:, 0] * 1e-6 * z_pu
    zk = np.sqrt(rk**2 + xk**2)
    vk_percent = np.sign(xk) * zk * 100
    vkr_percent = rk * 100
    pfe_kw = g0 * sn_mva * 1e3
    i0_percent = 100 * np.sqrt(b0**2 + g0**2) * net.sn_mva / sn_mva
    taps = data[key].loc[:, ("Phase Shifting Properties", "Taps used for RAO")].str.split(
        ";", expand=True).astype(int).set_axis(["tap_min", "tap_max"], axis=1)

    du = _get_float_column(data[key], ("Phase Shifting Properties", "Phase Regulation δu (%)"))
    dphi = _get_float_column(data[key], ("Phase Shifting Properties", "Angle Regulation δu (%)"))
    phase_shifter = np.isclose(du, 0) & (~np.isclose(dphi, 0))  # Symmetrical/Asymmetrical not
    # considered

    _ = create_transformers_from_parameters(
        net,
        trafo_connections.hv_bus.values,
        trafo_connections.lv_bus.values,
        sn_mva,
        vn_hv_kv,
        vn_lv_kv,
        vkr_percent,
        vk_percent,
        pfe_kw,
        i0_percent,
        shift_degree=data[key].xs("Theta θ (°)", level=1, axis=1).values[:, 0],
        tap_pos=0,
        tap_neutral=0,
        tap_side="lv",
        tap_min=taps["tap_min"].values,
        tap_max=taps["tap_max"].values,
        tap_phase_shifter=phase_shifter,
        tap_step_percent=du,
        tap_step_degree=dphi,
        name=data[key].loc[:, ("Location", "Full Name")].str.strip().values,
        EIC_Code=data[key].xs("EIC_Code", level=1, axis=1).values[:, 0],
        TSO=data[key].xs("TSO", level=1, axis=1).values[:, 0],
        Comment=data[key].xs("Comment", level=1, axis=1).replace("\xa0", "").values[:, 0],
    )


def _invent_connections_between_grid_groups(
        net: pandapowerNet, minimal_trafo_invention: bool = False, **kwargs) -> None:
    """Adds connections between islanded grid groups via:

    - adding transformers between equally named buses that have different voltage level and lay in different groups
    - merge buses of same voltage level, different grid groups and equal name base
    - fuse buses that are close to each other

    Parameters
    ----------
    net : pandapowerNet
        net to be manipulated
    minimal_trafo_invention : bool, optional
        if True, adding transformers stops when no grid groups is islanded anymore (does not apply
        for release version 5 or 6, i.e. it does not care what value is passed to
        minimal_trafo_invention). If False, all equally named buses that have different voltage
        level and lay in different groups will be connected via additional transformers,
        by default False
    """
    grid_groups = get_grid_groups(net)
    bus_idx = _get_bus_idx(net)
    bus_grid_groups = pd.concat([pd.Series(group, index=buses) for group, buses in zip(
        grid_groups.index, grid_groups.buses)]).sort_index()

    # treat for example "Wuergau" equally as "Wuergau (2)":
    location_names = pd.Series(bus_idx.index.get_level_values(0))
    location_names = location_names.str.replace(r"(.) \([0-9]+\)", r"\1", regex=True)
    bus_idx.index = pd.MultiIndex.from_arrays(
        [location_names.values, bus_idx.index.get_level_values(1).to_numpy()],
        names=bus_idx.index.names)

    # --- add Transformers between equally named buses that have different voltage level and lay in
    # --- different groups
    connected_vn_kvs_by_trafos = pd.DataFrame({
        "hv": net.bus.vn_kv.loc[net.trafo.hv_bus.values].values,
        "lv": net.bus.vn_kv.loc[net.trafo.lv_bus.values].values,
        "index": net.trafo.index}).set_index(["hv", "lv"]).sort_index()
    dupl_location_names = location_names[location_names.duplicated()]

    for location_name in dupl_location_names:
        if minimal_trafo_invention and len(bus_grid_groups.unique()) <= 1:
            break  # break with regard to minimal_trafo_invention
        grid_groups_at_location = bus_grid_groups.loc[bus_idx.loc[location_name].values]
        grid_groups_at_location = grid_groups_at_location.drop_duplicates()
        if len(grid_groups_at_location) < 2:
            continue
        elif len(grid_groups_at_location) > 2:
            raise NotImplementedError("Code is not provided to invent Transformer connections "
                                      "between locations with more than two grid groups, i.e. "
                                      "voltage levels.")
        TSO = net.bus.zone.at[grid_groups_at_location.index[0]]
        vn_kvs = net.bus.vn_kv.loc[grid_groups_at_location.index].sort_values(ascending=False)
        try:
            trafos_connecting_same_voltage_levels = \
                connected_vn_kvs_by_trafos.loc[tuple(vn_kvs)]
        except KeyError:
            logger.info(f"For location {location_name}, no transformer data can be reused since "
                        f"no transformer connects {vn_kvs.sort_values(ascending=False).iat[0]} kV "
                        f"and {vn_kvs.sort_values(ascending=False).iat[1]} kV.")
            continue
        trafos_of_same_TSO = trafos_connecting_same_voltage_levels.loc[(net.bus.zone.loc[
            net.trafo.hv_bus.loc[trafos_connecting_same_voltage_levels.values.flatten(
            )].values] == TSO).values].values.flatten()

        # from which trafo parameters are copied:
        tr_to_be_copied = trafos_of_same_TSO[0] if len(trafos_of_same_TSO) else \
            trafos_connecting_same_voltage_levels.values.flatten()[0]

        # copy transformer data
        duplicated_row = net.trafo.loc[[tr_to_be_copied]].copy()
        duplicated_row.index = [net.trafo.index.max() + 1]  # adjust index
        duplicated_row.hv_bus = vn_kvs.index[0]  # adjust hv_bus, lv_bus
        duplicated_row.lv_bus = vn_kvs.index[1]  # adjust hv_bus, lv_bus
        duplicated_row.name = "additional transformer to connect the grid"
        net.trafo = pd.concat([net.trafo, duplicated_row])

        bus_grid_groups.loc[bus_grid_groups == grid_groups_at_location.iat[1]] = \
            grid_groups_at_location.iat[0]

    # --- merge buses of same voltage level, different grid groups and equal name base
    bus_name_splits = net.bus.name.str.split(r"[ -/]+", expand=True)
    buses_with_single_base = net.bus.name.loc[(~bus_name_splits.isnull()).sum(axis=1) == 1]
    for idx, name_base in buses_with_single_base.items():
        same_name_base = net.bus.drop(idx).name.str.contains(name_base)
        if not any(same_name_base):
            continue
        other_group = bus_grid_groups.drop(idx) != bus_grid_groups.at[idx]
        same_vn = net.bus.drop(idx).vn_kv == net.bus.vn_kv.at[idx]
        is_fuse_candidate = same_name_base & other_group & same_vn
        if not any(is_fuse_candidate):
            continue
        to_fuse = bus_grid_groups.drop(idx).loc[is_fuse_candidate].drop_duplicates()
        fuse_buses(net, idx, set(to_fuse.index))

        bus_grid_groups.loc[bus_grid_groups.isin(bus_grid_groups.drop(idx).loc[
            is_fuse_candidate].unique())] = grid_groups_at_location.iat[0]
        bus_grid_groups = bus_grid_groups.drop(to_fuse.index)

    # --- fuse buses that are close to each other
    for name1, name2 in [("CROISIERE", "BOLLENE (POSTE RESEAU)"),
                         ("CAEN", "DRONNIERE (LA)"),
                         ("TRINITE-VICTOR", "MENTON/TRINITE VICTOR")]:
        b1 = net.bus.index[net.bus.name == name1]
        b2 = net.bus.index[net.bus.name == name2]
        if len(b1) == 1 and len(b2) >= 1:
            fuse_buses(net, b1[0], set(b2))
            bus_grid_groups = bus_grid_groups.drop(b2)
        else:
            logger.info("Buses of the following names were intended to be fused but were not found."
                        f"\n'{name1}' and '{name2}'")


def drop_islanded_grid_groups(
        net: pandapowerNet,
        min_bus_number: Union[int, str],
        **kwargs) -> None:
    """Drops grid groups that are islanded and include a number of buses below min_bus_number.

    Parameters
    ----------
    net : pandapowerNet
        net in which islanded grid groups will be dropped
    min_bus_number : int | str, optional
        Threshold value to decide which small grid groups should be dropped and which large grid
        groups should be kept. If all islanded grid groups should be dropped except of the one
        largest, set "max". If all grid groups that do not contain a slack element should be
        dropped, set "unsupplied".
    """
    def _grid_groups_to_drop_by_min_bus_number():
        return grid_groups.loc[grid_groups["n_buses"] < min_bus_number]

    grid_groups = get_grid_groups(net, **kwargs)

    if min_bus_number == "unsupplied":
        slack_buses = set(net.ext_grid.loc[net.ext_grid.in_service, "bus"]) | \
            set(net.gen.loc[net.gen.in_service & net.gen.slack, "bus"])
        grid_groups_to_drop = grid_groups.loc[~grid_groups.buses.apply(
            lambda x: not x.isdisjoint(slack_buses))]

    elif min_bus_number == "max":
        min_bus_number = grid_groups["n_buses"].max()
        grid_groups_to_drop = _grid_groups_to_drop_by_min_bus_number()

    elif isinstance(min_bus_number, int):
        grid_groups_to_drop = _grid_groups_to_drop_by_min_bus_number()

    else:
        raise NotImplementedError(
            f"{min_bus_number=} is not implemented. Use an int, 'max', or 'unsupplied' instead.")

    buses_to_drop = reduce(set.union, grid_groups_to_drop.buses)
    drop_buses(net, buses_to_drop)
    logger.info(f"drop_islanded_grid_groups() drops {len(grid_groups_to_drop)} grid groups with a "
                f"total of {grid_groups_to_drop.n_buses.sum()} buses.")


def _add_bus_geo(net: pandapowerNet, line_geo_data: pd.DataFrame) -> None:
    """Adds geodata to the buses. The function needs to handle cases where line_geo_data does not
    include no or multiple geodata per bus. Primarly, the geodata are allocate via EIC Code names,
    if ambigous, names are considered.

    Parameters
    ----------
    net : pandapowerNet
        net in which geodata are added to the buses
    line_geo_data : pd.DataFrame
        Converted geodata from the html file
    """
    iSl = pd.IndexSlice
    lgd_EIC_bus = line_geo_data.pivot_table(values="value", index=["EIC_Code", "bus"],
                                            columns="geo_dim")
    lgd_name_bus = line_geo_data.pivot_table(values="value", index=["name", "bus"],
                                             columns="geo_dim")
    lgd_EIC_bus_idx_extended = pd.MultiIndex.from_frame(lgd_EIC_bus.index.to_frame().assign(
        **dict(col_name="EIC_Code")).rename(columns=dict(EIC_Code="identifier")).loc[
        :, ["col_name", "identifier", "bus"]])
    lgd_name_bus_idx_extended = pd.MultiIndex.from_frame(lgd_name_bus.index.to_frame().assign(
        **dict(col_name="name")).rename(columns=dict(name="identifier")).loc[
        :, ["col_name", "identifier", "bus"]])
    lgd_bus = pd.concat([lgd_EIC_bus.set_axis(lgd_EIC_bus_idx_extended),
                         lgd_name_bus.set_axis(lgd_name_bus_idx_extended)])
    dupl_EICs = net.line.EIC_Code.loc[net.line.EIC_Code.duplicated()]
    dupl_names = net.line.name.loc[net.line.name.duplicated()]

    def _geo_json_str(this_bus_geo: pd.Series) -> str:
        return f'{{"coordinates": [{this_bus_geo.at["lng"]}, {this_bus_geo.at["lat"]}], "type": "Point"}}'

    def _add_bus_geo_inner(bus: int) -> Optional[str]:
        from_bus_line_excerpt = net.line.loc[net.line.from_bus ==
                                             bus, ["EIC_Code", "name", "Tieline"]]
        to_bus_line_excerpt = net.line.loc[net.line.to_bus == bus, ["EIC_Code", "name", "Tieline"]]
        line_excerpt = pd.concat([from_bus_line_excerpt, to_bus_line_excerpt])
        n_connected_line_ends = len(line_excerpt)
        if n_connected_line_ends == 0:
            logger.error(
                f"Bus {bus} (name {net.bus.at[bus, 'name']}) is not found in line_geo_data.")
            return None
        is_dupl = pd.concat([
            pd.DataFrame({"EIC": from_bus_line_excerpt.EIC_Code.isin(dupl_EICs).values,
                          "name": from_bus_line_excerpt.name.isin(dupl_names).values},
                         index=pd.MultiIndex.from_product([["from"], from_bus_line_excerpt.index],
                                                          names=["bus", "line_index"])),
            pd.DataFrame({"EIC": to_bus_line_excerpt.EIC_Code.isin(dupl_EICs).values,
                          "name": to_bus_line_excerpt.name.isin(dupl_names).values},
                         index=pd.MultiIndex.from_product([["to"], to_bus_line_excerpt.index],
                                                          names=["bus", "line_index"]))
        ])
        is_missing = pd.DataFrame({
            "EIC": ~line_excerpt.EIC_Code.isin(
                lgd_bus.loc["EIC_Code"].index.get_level_values("identifier")),
            "name": ~line_excerpt.name.isin(
                lgd_bus.loc["name"].index.get_level_values("identifier"))
        }).set_axis(is_dupl.index)
        is_tieline = pd.Series(net.line.loc[is_dupl.index.get_level_values("line_index"),
                                            "Tieline"].values, index=is_dupl.index)

        # --- construct access_vals, i.e. values to take line geo data from lgd_bus
        # --- if not duplicated, take "EIC_Code". Otherwise and if not dupl, take "name".
        # --- Otherwise ignore. Do it for both from and to bus
        access_vals = pd.DataFrame({
            "col_name": "EIC_Code",
            "identifier": line_excerpt.EIC_Code.values,
            "bus": is_dupl.index.get_level_values("bus").values
        })  # default is EIC_Code
        take_from_name = ((is_dupl.EIC | is_missing.EIC) & (
            ~is_dupl.name & ~is_missing.name)).values
        access_vals.loc[take_from_name, "col_name"] = "name"
        access_vals.loc[take_from_name, "identifier"] = line_excerpt.name.loc[take_from_name].values
        keep = (~(is_dupl | is_missing)).any(axis=1).values
        if np.all(is_missing):
            log_msg = (f"For bus {bus} (name {net.bus.at[bus, 'name']}), {n_connected_line_ends} "
                       "were found but no EIC_Codes or names of corresponding lines were found ."
                       "in the geo data from the html file.")
            if is_tieline.all():
                logger.debug(log_msg)
            else:
                logger.warning(log_msg)
            return None
        elif sum(keep) == 0:
            logger.info(f"For {bus=}, all EIC_Codes and names of connected lines are ambiguous. "
                        "No geo data is dropped at this point.")
            keep[(~is_missing).any(axis=1)] = True
        access_vals = access_vals.loc[keep]

        # --- get this_bus_geo from EIC_Code or name with regard to access_vals
        this_bus_geo = lgd_bus.loc[iSl[
            access_vals.col_name, access_vals.identifier, access_vals.bus], :]

        if len(this_bus_geo) > 1:
            # reduce similar/equal lines
            this_bus_geo = this_bus_geo.loc[this_bus_geo.round(2).drop_duplicates().index]

        # --- return geo_json_str
        len_this_bus_geo = len(this_bus_geo)
        if len_this_bus_geo == 1:
            return _geo_json_str(this_bus_geo.iloc[0])
        elif len_this_bus_geo == 2:
            how_often = pd.Series(
                [sum(np.isclose(lgd_EIC_bus["lat"], this_bus_geo["lat"].iat[i]) &
                     np.isclose(lgd_EIC_bus["lng"], this_bus_geo["lng"].iat[i])) for i in
                 range(len_this_bus_geo)], index=this_bus_geo.index)
            if how_often.at[how_often.idxmax()] >= 1:
                logger.warning(f"Bus {bus} (name {net.bus.at[bus, 'name']}) was found multiple times"
                               " in line_geo_data. No value exists more often than others. "
                               "The first of most used geo positions is used.")
            return _geo_json_str(this_bus_geo.loc[how_often.idxmax()])

    net.bus.geo = [_add_bus_geo_inner(bus) for bus in net.bus.index]


# --- tertiary functions ---------------------------------------------------------------------------

def _float_col_comma_correction(data: dict[str, pd.DataFrame], key: str, col_names: list):
    for col_name in col_names:
        data[key][col_name] = pd.to_numeric(data[key][col_name].astype(str).str.replace(
            ",", "."), errors="coerce")


def _get_transformer_voltages(
        data: dict[str, pd.DataFrame], bus_idx: pd.Series) -> tuple[np.ndarray, np.ndarray]:

    key = "Transformers"
    vn = data[key].loc[:, [("Voltage_level(kV)", "Primary"),
                           ("Voltage_level(kV)", "Secondary")]].values
    vn_hv_kv = np.max(vn, axis=1)
    vn_lv_kv = np.min(vn, axis=1)
    if is_integer_dtype(list(bus_idx.index.dtypes)[1]):
        vn_hv_kv = vn_hv_kv.astype(int)
        vn_lv_kv = vn_lv_kv.astype(int)

    return vn_hv_kv, vn_lv_kv


def _allocate_trafos_to_buses_and_create_buses(
        net: pandapowerNet, data: dict[str, pd.DataFrame], bus_idx: pd.Series,
        vn_hv_kv: np.ndarray, vn_lv_kv: np.ndarray,
        rel_deviation_threshold_for_trafo_bus_creation: float = 0.2,
        log_rel_vn_deviation: float = 0.12, **kwargs) -> pd.DataFrame:
    """Provides a DataFrame of data to allocate transformers to the buses according to their
    location names. If locations of transformers do not exist due to the data of the lines and
    tielines sheets, additional buses are created. If locations exist but have a far different
    voltage level than the transformer, either a warning is logged or additional buses are created
    according to rel_deviation_threshold_for_trafo_bus_creation and log_rel_vn_deviation.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    data : dict[str, pd.DataFrame]
        _description_
    bus_idx : pd.Series
        Series of indices and corresponding location names and voltage levels in the MultiIndex of
        the Series
    vn_hv_kv : np.ndarray
        nominal voltages of the hv side of the transformers
    vn_lv_kv : np.ndarray
        Nominal voltages of the lv side of the transformers
    rel_deviation_threshold_for_trafo_bus_creation : float, optional
        If the voltage level of transformer locations is far different than the transformer data,
        additional buses are created. rel_deviation_threshold_for_trafo_bus_creation defines the
        tolerance in which no additional buses are created. By default 0.2
    log_rel_vn_deviation : float, optional
        This parameter allows a range below rel_deviation_threshold_for_trafo_bus_creation in which
        a warning is logged instead of a creating additional buses. By default 0.12

    Returns
    -------
    pd.DataFrame
        information to which bus the trafos should be connected to. Columns are
        ["name", "hv_bus", "lv_bus", "vn_hv_kv", "vn_lv_kv", ...]
    """

    if rel_deviation_threshold_for_trafo_bus_creation < log_rel_vn_deviation:
        logger.warning(
            f"Given parameters violates the ineqation "
            f"{rel_deviation_threshold_for_trafo_bus_creation=} >= {log_rel_vn_deviation=}. "
            f"Therefore, rel_deviation_threshold_for_trafo_bus_creation={log_rel_vn_deviation} "
            "is assumed.")
        rel_deviation_threshold_for_trafo_bus_creation = log_rel_vn_deviation

    key = "Transformers"
    bus_location_names = set(net.bus.name)
    trafo_bus_names = data[key].loc[:, ("Location", "Full Name")]
    trafo_location_names = _find_trafo_locations(trafo_bus_names, bus_location_names)

    # --- construct DataFrame trafo_connections including all information on trafo allocation to
    # --- buses
    empties = -1*np.ones(len(vn_hv_kv), dtype=int)
    trafo_connections = pd.DataFrame({
        "name": trafo_location_names,
        "hv_bus": empties,
        "lv_bus": empties,
        "vn_hv_kv": vn_hv_kv,
        "vn_lv_kv": vn_lv_kv,
        "vn_hv_kv_next_bus": vn_hv_kv,
        "vn_lv_kv_next_bus": vn_lv_kv,
        "hv_rel_deviation": np.zeros(len(vn_hv_kv)),
        "lv_rel_deviation": np.zeros(len(vn_hv_kv)),
    })
    trafo_connections[["hv_bus", "lv_bus"]] = trafo_connections[[
        "hv_bus", "lv_bus"]].astype(np.int64)

    for side in ["hv", "lv"]:
        bus_col, trafo_vn_col, next_col, rel_dev_col, has_dev_col = \
            f"{side}_bus", f"vn_{side}_kv", f"vn_{side}_kv_next_bus", f"{side}_rel_deviation", \
            f"trafo_{side}_to_bus_deviation"
        name_vn_series = pd.Series(
            tuple(zip(trafo_location_names, trafo_connections[trafo_vn_col])))
        isin = name_vn_series.isin(bus_idx.index)
        trafo_connections[has_dev_col] = ~isin
        trafo_connections.loc[isin, bus_col] = bus_idx.loc[name_vn_series.loc[isin]].values

        # --- code to find bus locations with vn deviation
        next_vn = np.array([bus_idx.loc[tln.name].index.values[
            (pd.Series(bus_idx.loc[tln.name].index) - getattr(tln, trafo_vn_col)).abs().idxmin(
            )] for tln in trafo_connections.loc[~isin, ["name", trafo_vn_col]].itertuples()])
        trafo_connections.loc[~isin, next_col] = next_vn
        rel_dev = np.abs(next_vn - trafo_connections.loc[~isin, trafo_vn_col].values) / next_vn
        trafo_connections.loc[~isin, rel_dev_col] = rel_dev
        trafo_connections.loc[~isin, bus_col] = \
            bus_idx.loc[list(tuple(zip(trafo_connections.loc[~isin, "name"],
                                       trafo_connections.loc[~isin, next_col])))].values

        # --- create buses to avoid too large vn deviations between nodes and transformers
        need_bus_creation = trafo_connections[rel_dev_col] > \
            rel_deviation_threshold_for_trafo_bus_creation
        new_bus_data = pd.DataFrame({
            "vn_kv": trafo_connections.loc[need_bus_creation, trafo_vn_col].values,
            "name": trafo_connections.loc[need_bus_creation, "name"].values,
            "TSO": data[key].loc[need_bus_creation, ("Location", "TSO")].values
        })
        new_bus_data_dd = _drop_duplicates_and_join_TSO(new_bus_data)
        new_bus_idx = create_buses(net, len(new_bus_data_dd), vn_kv=new_bus_data_dd.vn_kv,
                                   name=new_bus_data_dd.name, zone=new_bus_data_dd.TSO)
        trafo_connections.loc[need_bus_creation, bus_col] = net.bus.loc[new_bus_idx, [
            "name", "vn_kv"]].reset_index().set_index(["name", "vn_kv"]).loc[list(new_bus_data[[
                "name", "vn_kv"]].itertuples(index=False, name=None))].values
        trafo_connections.loc[need_bus_creation, next_col] = \
            trafo_connections.loc[need_bus_creation, trafo_vn_col].values
        trafo_connections.loc[need_bus_creation, rel_dev_col] = 0
        trafo_connections.loc[need_bus_creation, has_dev_col] = False

    # --- create buses for trafos that are connected to the same bus at both sides (possible if
    # --- vn_hv_kv < vn_lv_kv *(1+rel_deviation_threshold_for_trafo_bus_creation) which usually
    # --- occurs for PSTs only)
    same_bus_connection = trafo_connections.hv_bus == trafo_connections.lv_bus
    duplicated_buses = net.bus.loc[trafo_connections.loc[same_bus_connection, "lv_bus"]].copy()
    duplicated_buses["name"] += " (2)"
    duplicated_buses.index = list(range(net.bus.index.max()+1,
                                        net.bus.index.max()+1+len(duplicated_buses)))
    trafo_connections.loc[same_bus_connection, "lv_bus"] = duplicated_buses.index
    net.bus = pd.concat([net.bus, duplicated_buses])
    if n_add_buses := len(duplicated_buses):
        tr_names = data[key].loc[trafo_connections.index[same_bus_connection],
                                 ("Location", "Full Name")]
        are_PSTs = tr_names.str.contains("PST")
        logger.info(f"{n_add_buses} additional buses were created to avoid that transformers are "
                    f"connected to the same bus at both side, hv and lv. Of the causing "
                    f"{len(tr_names)} transformers, {sum(are_PSTs)} contain 'PST' in their name. "
                    f"According to this converter, the power flows over all these transformers will"
                    f" end at the additional buses. Please consider to connect lines with the "
                    f"additional buses, so that the power flow is over the (PST) transformers into "
                    f"the lines.")

    # --- log according to log_rel_vn_deviation
    for side in ["hv", "lv"]:
        need_logging = trafo_connections.loc[trafo_connections[has_dev_col],
                                             rel_dev_col] > log_rel_vn_deviation
        if n_need_logging := sum(need_logging):
            max_dev = trafo_connections[rel_dev_col].max()
            idx_max_dev = trafo_connections[rel_dev_col].idxmax()
            logger.warning(
                f"For {n_need_logging} Transformers ({side} side), only locations were found (orig"
                f"in are the line and tieline data) that have a higher relative deviation than "
                f"{log_rel_vn_deviation}. The maximum relative deviation is {max_dev} which "
                f"results from a Transformer rated voltage of "
                f"{trafo_connections.at[idx_max_dev, trafo_vn_col]} and a bus "
                f"rated voltage (taken from Lines/Tielines data sheet) of "
                f"{trafo_connections.at[idx_max_dev, next_col]}. The best locations were "
                f"nevertheless applied, due to {rel_deviation_threshold_for_trafo_bus_creation=}")

    assert (trafo_connections.hv_bus > -1).all()
    assert (trafo_connections.lv_bus > -1).all()
    assert (trafo_connections.hv_bus != trafo_connections.lv_bus).all()

    return trafo_connections


def _find_trafo_locations(trafo_bus_names, bus_location_names):
    # --- split (original and lower case) strings at " " separators to remove impeding parts for
    # identifying the location names
    trafo_bus_names_expended = trafo_bus_names.str.split(r"[ ]+|-A[0-9]+|-TD[0-9]+|-PF[0-9]+",
                                                         expand=True).fillna("").replace(" ", "")
    trafo_bus_names_expended_lower = trafo_bus_names.str.lower().str.split(
        r"[ ]+|-A[0-9]+|-TD[0-9]+|-PF[0-9]+", expand=True).fillna("").replace(" ", "")

    # --- identify impeding parts
    contains_number = trafo_bus_names_expended.map(lambda x: any(char.isdigit() for char in x))
    to_drop = (trafo_bus_names_expended_lower == "tr") | (trafo_bus_names_expended_lower == "pst") \
        | (trafo_bus_names_expended == "") | (trafo_bus_names_expended == "/") | (
        trafo_bus_names_expended == "LIPST") | (trafo_bus_names_expended == "EHPST") | (
        trafo_bus_names_expended == "TFO") | (trafo_bus_names_expended_lower == "trafo") | (
        trafo_bus_names_expended_lower == "kv") | contains_number
    trafo_bus_names_expended[to_drop] = ""

    # --- reconstruct name strings for identification
    trafo_bus_names_joined = trafo_bus_names_expended.where(~to_drop).fillna('').agg(
        ' '.join, axis=1).str.strip()
    trafo_bus_names_longest_part = trafo_bus_names_expended.apply(
        lambda row: max(row, key=len), axis=1)
    joined_in_buses = trafo_bus_names_joined.isin(bus_location_names)
    longest_part_in_buses = trafo_bus_names_longest_part.isin(bus_location_names)

    # --- check whether all name strings point at location names of the buses
    if False:  # for easy testing
        fail = ~(joined_in_buses | longest_part_in_buses)
        a = pd.concat([trafo_bus_names_joined.loc[fail],
                      trafo_bus_names_longest_part.loc[fail]], axis=1)

    if n_bus_names_not_found := len(joined_in_buses) - sum(joined_in_buses | longest_part_in_buses):
        raise ValueError(
            f"For {n_bus_names_not_found} Tranformers, no suitable bus location names were found, "
            f"i.e. the algorithm did not find a (part) of Transformers-Location-Full Name that fits"
            " to Substation_1 or Substation_2 data in Lines or Tielines sheet.")

    # --- set the trafo location names and trafo bus indices respectively
    trafo_location_names = trafo_bus_names_longest_part
    trafo_location_names.loc[joined_in_buses] = trafo_bus_names_joined

    return trafo_location_names


def _drop_duplicates_and_join_TSO(bus_df: pd.DataFrame) -> pd.DataFrame:
    bus_df = bus_df.drop_duplicates(ignore_index=True)
    # just keep one bus per name and vn_kv. If there are multiple buses of different TSOs, join the
    # TSO strings:
    bus_df = bus_df.groupby(["name", "vn_kv"], as_index=False).agg({"TSO": lambda x: '/'.join(x)})
    assert not bus_df.duplicated(["name", "vn_kv"]).any()
    return bus_df


def _get_float_column(df, col_tuple, fill=0):
    series = df.loc[:, col_tuple]
    series.loc[series == "\xa0"] = fill
    return series.astype(float).fillna(fill)


def _get_bus_idx(net: pandapowerNet) -> pd.Series:
    return net.bus[["name", "vn_kv"]].rename_axis("index").reset_index().set_index([
        "name", "vn_kv"])["index"]


def get_grid_groups(net: pandapowerNet, **kwargs) -> pd.DataFrame:
    notravbuses_dict = dict() if "notravbuses" not in kwargs.keys() else {
        "notravbuses": kwargs.pop("notravbuses")}
    grid_group_buses = [set_ for set_ in connected_components(create_nxgraph(net, **kwargs),
                                                              **notravbuses_dict)]
    grid_groups = pd.DataFrame({"buses": grid_group_buses})
    grid_groups["n_buses"] = grid_groups["buses"].apply(len)
    return grid_groups


def _lng_lat_to_df(dict_: dict, line_EIC: str, line_name: str) -> pd.DataFrame:
    return pd.DataFrame([
        [line_EIC, line_name, "from", "lng", dict_["lng"][0]],
        [line_EIC, line_name,   "to", "lng", dict_["lng"][1]],
        [line_EIC, line_name, "from", "lat", dict_["lat"][0]],
        [line_EIC, line_name,   "to", "lat", dict_["lat"][1]],
    ], columns=["EIC_Code", "name", "bus", "geo_dim", "value"])


def _fill_geo_at_one_sided_branches_without_geo_extent(net: pandapowerNet):

    def _check_geo_availablitiy(net: pandapowerNet) -> dict[str, Union[pd.Index, int]]:
        av = dict()  # availablitiy of geodata
        av["bus_with_geo"] = net.bus.index[~net.bus.geo.isnull()]
        av["lines_fbw_tbwo"] = net.line.index[net.line.from_bus.isin(av["bus_with_geo"]) &
                                              (~net.line.to_bus.isin(av["bus_with_geo"]))]
        av["lines_fbwo_tbw"] = net.line.index[(~net.line.from_bus.isin(av["bus_with_geo"])) &
                                              net.line.to_bus.isin(av["bus_with_geo"])]
        av["trafos_hvbw_lvbwo"] = net.trafo.index[net.trafo.hv_bus.isin(av["bus_with_geo"]) &
                                                  (~net.trafo.lv_bus.isin(av["bus_with_geo"]))]
        av["trafos_hvbwo_lvbw"] = net.trafo.index[(~net.trafo.hv_bus.isin(av["bus_with_geo"])) &
                                                  net.trafo.lv_bus.isin(av["bus_with_geo"])]
        av["n_lines_one_side_geo"] = len(av["lines_fbw_tbwo"])+len(av["lines_fbwo_tbw"])
        return av

    geo_avail = _check_geo_availablitiy(net)
    while geo_avail["n_lines_one_side_geo"]:

        # copy available geodata to the other end of branches where geodata are missing
        for et, bus_w_geo, bus_wo_geo, idx_key in zip(
                ["line", "line", "trafo", "trafo"],
                ["to_bus", "from_bus", "lv_bus", "hv_bus"],
                ["from_bus", "to_bus", "hv_bus", "lv_bus"],
                ["lines_fbwo_tbw", "lines_fbw_tbwo", "trafos_hvbwo_lvbw", "trafos_hvbw_lvbwo"]):
            net.bus.loc[net[et].loc[geo_avail[idx_key], bus_wo_geo].values, "geo"] = \
                net.bus.loc[net[et].loc[geo_avail[idx_key], bus_w_geo].values, "geo"].values
        geo_avail = _check_geo_availablitiy(net)

    set_line_geodata_from_bus_geodata(net)


def _multi_str_repl(st: str, repl: list[tuple]) -> str:
    for (old, new) in repl:
        return st.replace(old, new)


if __name__ == "__main__":
    from pathlib import Path
    import os
    import pandapower as pp

    home = str(Path.home())
    jao_data_folder = os.path.join(home, "Documents", "JAO Static Grid Model")

    release5 = os.path.join(jao_data_folder, "20240329_Core Static Grid Model – 5th release")
    excel_file_path = os.path.join(release5, "20240329_Core Static Grid Model_public.xlsx")
    html_file_path = os.path.join(release5, "20240329_Core Static Grid Model Map_public",
                                  "2024-03-18_Core_SGM_publication.html")

    release6 = os.path.join(jao_data_folder, "202409_Core Static Grid Mode_6th release")
    excel_file_path = os.path.join(release6, "20240916_Core Static Grid Model_for publication.xlsx")
    html_file_path = os.path.join(release6, "2024-09-13_Core_SGM_publication_files",
                                  "2024-09-13_Core_SGM_publication.html")

    pp_net_json_file = os.path.join(home, "desktop", "jao_grid.json")

    if 1:  # read from original data
        net = from_jao(excel_file_path, html_file_path, True, drop_grid_groups_islands=True)
        pp.to_json(net, pp_net_json_file)
    else:  # load net from already converted and stored net
        net = pp.from_json(pp_net_json_file)

    print(net)
    grid_groups = get_grid_groups(net)
    print(grid_groups)

    _fill_geo_at_one_sided_branches_without_geo_extent(net)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from pandapower.converter.pandamodels.to_pm import *


from cmath import rect

from numpy import real, vectorize, deg2rad, maximum, sqrt, empty, zeros, nan, int64

from pandapower.pypower.idx_brch import F_BUS, T_BUS
from pandapower.pf.pfsoln_numba import calc_branch_flows_batch
from pandapower.pypower.idx_bus import BASE_KV
from pandapower.results_branch import _get_trafo3w_lookups
from pandapower.results_bus import _get_bus_idx


def get_batch_bus_results(net, vm, va):
    # convert ppci bus results to net.res_bus results
    # we need va, vm with out of service elements
    # create matrix of proper size
    bus_shape_ppc = net["_ppc"]["bus"].shape[0]
    bus_idx = _get_bus_idx(net)
    # (n_ts, n_bus in ppc)
    vm_full = empty((vm.shape[0], bus_shape_ppc))
    va_full = empty((va.shape[0], bus_shape_ppc))
    vm_full.fill(nan)
    va_full.fill(nan)
    # and fill it
    vm_full[:, :vm.shape[1]] = vm.values
    va_full[:, :va.shape[1]] = va.values
    return vm_full[:, bus_idx], va_full[:, bus_idx]


def get_batch_line_results(net, i_abs):
    f, t = net._pd2ppc_lookups["branch"]["line"]
    line_df = net["line"]
    i_max = line_df["max_i_ka"].values * line_df["df"].values * line_df["parallel"].values
    i_ka = maximum(i_abs[0][:, f:t], i_abs[1][:, f:t])
    loading_percent = i_ka / i_max * 100
    i_from_ka = i_abs[0][:, :len(net.line)]
    i_to_ka = i_abs[1][:, :len(net.line)]
    return i_ka, i_from_ka, i_to_ka, loading_percent


def get_batch_trafo_results(net, i_abs, s_abs):
    if "trafo" not in net._pd2ppc_lookups["branch"]:
        return
    f, t = net._pd2ppc_lookups["branch"]["trafo"]

    i_ka = maximum(i_abs[0][:, f:t], i_abs[1][:, f:t])
    s_mva = maximum(s_abs[0][:, f:t], s_abs[1][:, f:t])
    sn_mva = net["trafo"]["sn_mva"].values

    trafo_loading = net["_options"]["trafo_loading"]

    if trafo_loading == "current":
        # get loading percent from rated current
        trafo_df = net["trafo"]
        s_hv = i_abs[0][:, f:t] * trafo_df["vn_hv_kv"].values * sqrt(3) / sn_mva * 100.
        s_lv = i_abs[1][:, f:t] * trafo_df["vn_lv_kv"].values * sqrt(3) / sn_mva * 100.
        ld_trafo = maximum(s_hv, s_lv)
    elif trafo_loading == "power":
        # get loading percent from rated power
        ld_trafo = s_mva / sn_mva * 100.
    else:
        raise ValueError(
            "Unknown transformer loading parameter %s - choose 'current' or 'power'" % trafo_loading)

    loading_percent = ld_trafo / net["trafo"]["parallel"].values / net["trafo"]["df"].values
    i_hv_ka = i_abs[0][:, f:t]
    i_lv_ka = i_abs[1][:, f:t]
    return i_ka, i_hv_ka, i_lv_ka, s_mva, loading_percent


def get_batch_trafo3w_results(net, i_abs, s_abs):
    if "trafo3w" not in net._pd2ppc_lookups["branch"]:
        return

    f, hv, mv, lv = _get_trafo3w_lookups(net)
    i_h = i_abs[0][:, f:hv]
    i_m = i_abs[1][:, hv:mv]
    i_l = i_abs[1][:, mv:lv]

    t3 = net["trafo3w"]
    trafo_loading = net["_options"]["trafo_loading"]
    if trafo_loading == "current":
        ld_h = i_h * t3["vn_hv_kv"].values * sqrt(3) / t3["sn_hv_mva"].values * 100
        ld_m = i_m * t3["vn_mv_kv"].values * sqrt(3) / t3["sn_mv_mva"].values * 100
        ld_l = i_l * t3["vn_lv_kv"].values * sqrt(3) / t3["sn_lv_mva"].values * 100
        ld_trafo = maximum(maximum(ld_h, ld_m), ld_l)
    elif trafo_loading == "power":
        ld_h = s_abs[0][:, f:hv] / t3["sn_hv_mva"].values * 100.
        ld_m = s_abs[1][:, hv:mv] / t3["sn_mv_mva"].values * 100.
        ld_l = s_abs[1][:, mv:lv] / t3["sn_lv_mva"].values * 100.
        ld_trafo = maximum(maximum(ld_h, ld_m), ld_l)
    else:
        raise ValueError(
            "Unknown transformer loading parameter %s - choose 'current' or 'power'" % trafo_loading)

    return i_h, i_m, i_l, ld_trafo


def _get_empty_branch(ppc_branch_shape):
    empties = (empty(ppc_branch_shape, dtype=complex), empty(ppc_branch_shape, dtype=complex),
               empty(ppc_branch_shape), empty(ppc_branch_shape), empty(ppc_branch_shape), empty(ppc_branch_shape))
    for e in empties:
        e.fill(nan)
    return empties


def v_to_i_s(net, vm, va):
    ppc = net["_ppc"]
    internal = ppc["internal"]
    Yf = internal["Yf"]
    Yt = internal["Yt"]
    V = polar_to_rad(vm, va)

    baseMVA = internal["baseMVA"]
    branch = internal["branch"]
    base_kv = internal["bus"][:, BASE_KV]
    f_bus = real(branch[:, F_BUS]).astype(int64)
    t_bus = real(branch[:, T_BUS]).astype(int64)

    # batch read
    Sb_f, sf_abs, if_abs = calc_branch_flows_batch(Yf.data, Yf.indptr, Yf.indices, V, baseMVA, Yf.shape[0],
                                                   f_bus, base_kv)
    Sb_t, st_abs, it_abs = calc_branch_flows_batch(Yt.data, Yt.indptr, Yt.indices, V, baseMVA, Yt.shape[0],
                                                   t_bus, base_kv)

    # get ppc shaped values
    ppc = net["_ppc"]
    ppc_branch_shape = (V.shape[0], ppc["branch"].shape[0])
    sb_f_ppc, sb_t_ppc, s_f_abs_ppc, s_t_abs_ppc, i_f_abs_ppc, i_t_abs_ppc = _get_empty_branch(ppc_branch_shape)

    in_service = ppc["internal"]['branch_is']
    sb_f_ppc[:, in_service] = Sb_f
    sb_t_ppc[:, in_service] = Sb_t
    s_f_abs_ppc[:, in_service] = sf_abs
    s_t_abs_ppc[:, in_service] = st_abs
    i_f_abs_ppc[:, in_service] = if_abs
    i_t_abs_ppc[:, in_service] = it_abs

    return (sb_f_ppc, sb_t_ppc), (s_f_abs_ppc, s_t_abs_ppc), (i_f_abs_ppc, i_t_abs_ppc)


def polar_to_rad(vm, va):
    # get complex V matrix (input to batch branch flow function) from vm and va matrices
    nprect = vectorize(rect)
    return nprect(vm, deg2rad(va))



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
from pandapower.auxiliary import _preserve_dtypes

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)

import numpy as np
import pandas as pd
import networkx as nx
from copy import deepcopy

from pandapower.create import create_bus
from pandapower.run import rundcpp
from pandapower.topology.graph_searches import connected_component
from pandapower.shortcircuit.calc_sc import calc_sc
from pandapower.create import _get_index_with_check
from pandapower.topology import create_nxgraph
from pandapower.pypower.idx_bus import BUS_I
from pandapower.pypower.idx_brch import F_BUS, T_BUS, TAP

__all__ = ["detect_power_station_unit", "calc_sc_on_line"]

def detect_power_station_unit(net, mode="auto",
                              max_gen_voltage_kv=80, max_distance_km=0.01):
    """
    Identifies the power station units configuration (gen and trafo) according to IEC 60909.
    Updates the power_station_trafo in the gen table

    INPUT:
        **net** - panpdapower net

        **mode** (str, ("auto", trafo""))

        **max_gen_voltage_level** (float)

        **max_distance_km** (float)
    """
    logger.info("This function will overwrites the value 'power_station_trafo' in gen table")
    net.gen["power_station_trafo"] = np.nan

    required_gen = net.gen.loc[net.bus.loc[net.gen.bus.values, "vn_kv"].values < max_gen_voltage_kv,:]
    required_gen_bus = required_gen.loc[:, "bus"].values

    if mode.lower() == "auto":
        required_trafo = net.trafo.loc[net.bus.loc[net.trafo.lv_bus.values, "vn_kv"].values < max_gen_voltage_kv, :]
    elif mode.lower() == "trafo":
        if "power_station_unit" in net.trafo.columns:
            required_trafo = net.trafo.loc[net.trafo.power_station_unit, :]
        else:
            logger.warning("Using mode 'trafo' requires 'power_station_unit' defined for trafo! Using 'auto' mode instead!")
            required_trafo = net.trafo.loc[net.bus.loc[net.trafo.lv_bus.values, "vn_kv"].values < max_gen_voltage_kv, :]

    else:
        raise UserWarning(f"Unsupported modes: {mode}")

    trafo_lv_bus = net.trafo.loc[required_trafo.index, "lv_bus"].values
    trafo_hv_bus = net.trafo.loc[required_trafo.index, "hv_bus"].values

    g = create_nxgraph(net, respect_switches=True,
                       nogobuses=None, notravbuses=trafo_hv_bus)

    for t_ix in required_trafo.index:
        t_lv_bus = required_trafo.at[t_ix, "lv_bus"]
        bus_dist = pd.Series(nx.single_source_dijkstra_path_length(g, t_lv_bus, weight='weight'))

        connected_bus_at_lv_side = bus_dist[bus_dist < max_distance_km].index.values
        gen_bus_at_lv_side = np.intersect1d(connected_bus_at_lv_side, required_gen_bus)

        if len(gen_bus_at_lv_side) == 1:
            # Check parallel trafo
            if not len(np.intersect1d(connected_bus_at_lv_side, trafo_lv_bus)) == 1:
                raise UserWarning("Failure in power station units detection! Parallel trafos on generator detected!")
            if np.isin(required_gen_bus, gen_bus_at_lv_side).sum() > 1:
                logger.info("More than 1 gen detected at the lv side of a power station trafo! Will not be considered as power station unit")
                continue
            net.gen.loc[np.isin(net.gen.bus.values, gen_bus_at_lv_side),
                        "power_station_trafo"] = t_ix


def _create_element_from_exisiting(net, ele_type, ele_ix):
    dtypes = net[ele_type].dtypes
    ps = pd.Series(net[ele_type].loc[ele_ix, :].to_dict(), name=_get_index_with_check(net, ele_type,
                                                                                      None))

    net[ele_type] = pd.concat([net[ele_type], ps.to_frame().T])
    _preserve_dtypes(net[ele_type], dtypes)

    return net[ele_type].index.to_numpy()[-1]


def _create_aux_net(net, line_ix, distance_to_bus0):
    if distance_to_bus0 < 0 or distance_to_bus0 > 1:
        raise UserWarning("Calculating SC current on line failed! distance_to_bus0 must be between 0-1!")
    aux_net = deepcopy(net)

    # Create auxiliary bus
    aux_bus = create_bus(aux_net, vn_kv=aux_net.bus.at[aux_net.line.at[line_ix, "from_bus"], "vn_kv"],
                            name="aux_bus_sc_calc")

    # Create auxiliary line, while preserve the original index
    aux_line0 = _create_element_from_exisiting(aux_net, "line", line_ix)
    aux_line1 = _create_element_from_exisiting(aux_net, "line", line_ix)

    ## Update distance and auxiliary bus
    aux_net.line.at[aux_line0, "length_km"] = distance_to_bus0 * aux_net.line.at[line_ix, "length_km"]
    aux_net.line.at[aux_line0, "to_bus"] = aux_bus
    aux_net.line.at[aux_line0, "name"] += "_aux_line0"

    aux_net.line.at[aux_line1, "length_km"] = (1 - distance_to_bus0) * aux_net.line.at[line_ix, "length_km"]
    aux_net.line.at[aux_line1, "from_bus"] = aux_bus
    aux_net.line.at[aux_line1, "name"] += "_aux_line1"

    ## Disable original line
    aux_net.line.at[line_ix, "in_service"] = False

    ## Update line switch
    for switch_ix in aux_net.switch.query(f" et == 'l' and element == {line_ix}").index:
        aux_switch_ix = _create_element_from_exisiting(aux_net, "switch", switch_ix)
        if aux_net.switch.at[aux_switch_ix, "bus"] == aux_net.line.at[line_ix, "from_bus"]:
            # The from side switch connected to aux_line0
            aux_net.switch.at[aux_switch_ix, "element"] =  aux_line0
        else:
            # The to side switch connected to aux_line1
            aux_net.switch.at[aux_switch_ix, "element"] =  aux_line1
    return aux_net, aux_bus


def calc_sc_on_line(net, line_ix, distance_to_bus0, **kwargs):
    """
    Calculate the shortcircuit in the middle of the line, returns a modified network
    with the shortcircuit calculation results and the bus added

    INPUT:
        **net** - panpdapower net

        **line_ix** (int) - The line of the shortcircuit

        **distance_to_bus0** (float) - The position of the shortcircuit should be between 0-1

    OPTIONAL:
        **kwargs**** - the parameters required for the pandapower calc_sc function
    """
    # Update network
    aux_net, aux_bus = _create_aux_net(net, line_ix, distance_to_bus0)

    rundcpp(aux_net)
    calc_sc(aux_net, bus=aux_bus, **kwargs)

    # Return the new net and the aux bus
    return aux_net, aux_bus


def adjust_V0_for_trafo_tap(ppci, V0, bus_idx):
    branch = ppci["branch"]
    tap = ppci["branch"][:, TAP]
    tap_branch_idx = np.flatnonzero(tap != 1)

    if len(tap_branch_idx) == 0:
        return

    Zbus = ppci["internal"]["Zbus"]
    bus = ppci["bus"]

    f = (branch[:, F_BUS]).real.astype(np.int64)  ## list of "from" buses
    t = (branch[:, T_BUS]).real.astype(np.int64)  ## list of "to" buses
    nl = len(branch)
    nb = len(bus)

    mg = nx.MultiGraph()
    mg.add_nodes_from(bus[:, BUS_I].astype(np.int64).tolist())
    mg.add_edges_from(branch[:, [F_BUS, T_BUS]].real.astype(np.int64).tolist())

    hv_buses = branch[tap_branch_idx, F_BUS].real.astype(np.int64)
    lv_buses = branch[tap_branch_idx, T_BUS].real.astype(np.int64)

    for bh, bl, t in zip(hv_buses, lv_buses, tap[tap_branch_idx]):
        c = connected_component(mg, bh, notravbuses={bl})
        c = [cc for cc in c if cc != bl]

        if not np.intersect1d(c, bus_idx):
            for b in c:
                V0[b] = (Zbus[:, b] / Zbus[b, b] * V0[b] * t)[b]
        else:
            c = connected_component(mg, bl, notravbuses={bh})
            c = [cc for cc in c if cc != bh]
            if not np.intersect1d(c, bus_idx):
                for b in c:
                    V0[b] = (Zbus[:, b] / Zbus[b, b] * V0[b] / t)[b]

        # V0[c] = (Zbus[:, c] / Zbus[c, c] * V0[c] * tap[tap_branch_idx])[c]

    ## build connection matrix Cft = Cf - Ct for line and from - to buses
    # i = np.r_[range(nl), range(nl)]  ## double set of row indices
    ## connection matrix
    # Cft = sparse((np.r_[np.ones(nl), -np.ones(nl)], (i, np.r_[f, t])), (nl, nb))


def adjacency(ppci):
    branch = ppci["branch"]
    bus = ppci["bus"]
    tap = ppci["branch"][:, TAP].real
    tap_branch_idx = np.flatnonzero(tap != 1)
    all_vertices = bus[:, BUS_I]
    num_vertices = len(bus)
    f = (branch[:, F_BUS]).real.astype(np.int64)  ## list of "from" buses
    t = (branch[:, T_BUS]).real.astype(np.int64)  ## list of "to" buses

    hv_buses = branch[tap_branch_idx, F_BUS].real.astype(np.int64)
    lv_buses = branch[tap_branch_idx, T_BUS].real.astype(np.int64)

    matrix = np.zeros((num_vertices, num_vertices))
    matrix[f, t] = -1
    matrix[t, f] = -1
    matrix[np.diag_indices(num_vertices)] = -matrix.sum(axis=1)

    # matrix[hv_buses, lv_buses] = 1 / tap[tap_branch_idx]
    # matrix[lv_buses, hv_buses] = 1 / tap[tap_branch_idx]
    # matrix[hv_buses, hv_buses] = 1 / np.square(tap[tap_branch_idx])
    #
    # init = np.ones(num_vertices)
    # init[4] = 1.1
    # fake_i = matrix.dot(init)
    # print(fake_i)
    # np.flatnonzero(fake_i)

    return matrix



import os
import tkinter as tk
from pandapower.auxiliary import ADict
from pandapower.converter.powerfactory.main_pf import exit_gracefully

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def cancel(app, input_panel):
    logger.debug('received a cancel request from the user')
    input_panel.destroy()
    logger.debug('destroyed input panel, will attempt exit()')
    for h in logger.handlers:
        logger.removeHandler(h)
    exit_gracefully(app, 'exiting script', False)


def calc_test(app, **kwargs):
    logger.info('TESTING')
    exit_gracefully(app, 'test complete', False)


def browse_dst_test(input_panel, entry_path_dst):
    logger.info('BROWSE DESTINATION')
    entry_path_dst.delete(0, tk.END)
    entry_path_dst.insert(tk.END, 'HELLO')


def conf_var_graphics(app, opt_grf, var_graphics):
    logger.debug('configuring options for graphic options menu')
    dia_folder = app.GetProjectFolder('dia')
    dia_grid = [d.loc_name for d in dia_folder.GetContents('*.IntGrfnet')]
    dia_grid.sort()
    logger.info('found network diagrams in project: %s' % dia_grid)
    menu_options = ['no geodata', 'GPS']
    menu_options.extend(dia_grid)

    opt_grf['menu'].delete(0, 'end')
    for option in menu_options:
        opt_grf['menu'].add_command(label=option, command=tk._setit(var_graphics, option))


def make_gui(app, project_name, browse_dst, calc):
    params = ADict()
    working_directory = os.getcwd()
    input_panel = tk.Tk()
    params.input_panel = input_panel
    input_panel.iconbitmap(working_directory + r'\power_factory_files\favicon.ico')

    input_panel.title('Pandapower export')

    # row 0
    label_path_dst = tk.Label(input_panel, anchor='w', text='Path to the destination directory '
                                                            'for pandapower export: ')
    label_path_dst.grid(row=0, column=0, columnspan=3, sticky='w', padx=2, pady=4)

    entry_path_dst = tk.Entry(input_panel, width=70)
    entry_path_dst.delete(0, tk.END)
    entry_path_dst.grid(row=0, column=4, columnspan=3, padx=2, pady=2)
    # entry_path_dst.insert(0, r'C:\pp_projects\test') ##for testing

    params.entry_path_dst = entry_path_dst

    path_dst_button = tk.Button(input_panel, text='Browse', width=8,
                                command=lambda: browse_dst(input_panel, entry_path_dst))
    path_dst_button.grid(row=0, column=7, padx=4, pady=4)

    # row 1
    label_fname = tk.Label(input_panel, anchor='w', text='Provide file name for the network: ')
    label_fname.grid(row=1, column=0, columnspan=3, sticky='w', padx=2, pady=2)

    entry_fname = tk.Entry(input_panel, width=70)
    entry_fname.delete(0, tk.END)
    entry_fname.grid(row=1, column=4, columnspan=3, padx=2, pady=0)
    entry_fname.insert(0, project_name)
    # entry_fname.insert(0, 'test') ##for testing
    params.entry_fname = entry_fname

    # row 2 col 0-1
    L_OPTIONS = ('plini', 'plini_a', 'm:P:bus1')
    G_OPTIONS = ('pgini', 'pgini_a', 'm:P:bus1')
    GRF_OPTIONS = ('no geodata', 'GPS', 'graphic objects')
    US_OPTIONS = ('Deactivate', 'Drop', 'Nothing')
    SAVE_OPTIONS = ("JSON", "Excel")
    TAP_OPTIONS = ("nntap", "c:nntap")

    # PowerFactory variable to be exported as power value for loads
    var_p_loads = tk.StringVar(input_panel)
    var_p_loads.set(L_OPTIONS[0])
    tk.Label(input_panel, anchor='w', text='Loads P variable:').grid(row=2, column=0,
                                                                     sticky=tk.W, pady=0)
    opt_l = tk.OptionMenu(input_panel, var_p_loads, *L_OPTIONS)
    opt_l.grid(row=2, column=1, sticky="ew", pady=0)
    opt_l.config(width=10)

    params.pf_variable_p_loads = var_p_loads.get

    # row 3 col 0-1
    var_p_gen = tk.StringVar(input_panel)
    var_p_gen.set(G_OPTIONS[0])
    tk.Label(input_panel, anchor='w', text='Generators P variable:').grid(row=3, column=0,
                                                                          sticky=tk.W,
                                                                          pady=0)

    opt_g = tk.OptionMenu(input_panel, var_p_gen, *G_OPTIONS)
    opt_g.grid(row=3, column=1, sticky="ew", pady=0)
    opt_g.config(width=10)

    params.pf_variable_p_gen = var_p_gen.get

    # row 4 col 0-1
    var_graphics = tk.StringVar(input_panel)
    var_graphics.set(GRF_OPTIONS[0])
    tk.Label(input_panel, anchor='w', text='Collect coordinates from:').grid(row=4, column=0,
                                                                             sticky=tk.W,
                                                                             pady=0)

    opt_grf = tk.OptionMenu(input_panel, var_graphics, *GRF_OPTIONS)
    opt_grf.grid(row=4, column=1, sticky="ew", pady=0)
    opt_grf.config(width=10)

   
    # refresh graphics option menu
    try:
        conf_var_graphics(app, opt_grf, var_graphics)
    except Exception as err:
        logger.error('could not find network diagrams: %s' % err)
        pass

    params.flag_graphics = var_graphics.get

    # row 5 col 0-1
    var_us = tk.StringVar(input_panel)
    var_us.set(US_OPTIONS[2])
    tk.Label(input_panel, anchor='w', text='Unsupplied Elements:').grid(row=5, column=0,
                                                                             sticky=tk.W,
                                                                             pady=0)

    opt_us = tk.OptionMenu(input_panel, var_us, *US_OPTIONS)
    opt_us.grid(row=5, column=1, sticky="ew", pady=0)
    opt_us.config(width=10)
    
    params.handle_us = var_us.get

    # row 6 col 0-1
    save_as = tk.StringVar(input_panel)
    save_as.set(SAVE_OPTIONS[0])
    tk.Label(input_panel, anchor='w', text='Save As:').grid(row=6, column=0,
                                                                             sticky=tk.W,
                                                                             pady=0)

    opt_save = tk.OptionMenu(input_panel, save_as, *SAVE_OPTIONS)
    opt_save.grid(row=6, column=1, sticky="ew", pady=0)
    opt_save.config(width=10)
    
    params.save_as = save_as.get

    # row 7 col 0-1
    tap_grf = tk.StringVar(input_panel)
    tap_grf.set(TAP_OPTIONS[0])
    tk.Label(input_panel, anchor='w', text='Tap:').grid(row=7, column=0, sticky=tk.W, pady=0)

    opt_tap = tk.OptionMenu(input_panel, tap_grf, *TAP_OPTIONS)
    opt_tap.grid(row=7, column=1, sticky="ew", pady=0)
    opt_tap.config(width=10)
    
    params.tap_opt = tap_grf.get

    # row 8
    max_iter = tk.IntVar(input_panel)
    max_iter.set(30)
    tk.Label(input_panel, anchor='w', text='Max. iterations:').grid(row=8, column=0, sticky=tk.W, pady=0)

    iter_entry = tk.Entry(input_panel, width=8)
    iter_entry.delete(0, tk.END)
    iter_entry.grid(row=8, column=1, padx=2, pady=2, sticky=tk.W)
    iter_entry.insert(0, max_iter.get())
    # entry_fname.insert(0, 'test') ##for testing
    params.max_iter_entry = iter_entry

    # row 2 col 2-3
    PV_SL = tk.IntVar()
    tk.Checkbutton(input_panel, text="Export 'PV' buses as slack buses", variable=PV_SL).grid(row=2,
                                                                                              column=4,
                                                                                              sticky=tk.W)
    params.pv_as_slack = PV_SL.get


    EXPORT_CONTROLLER = tk.IntVar()
    EXPORT_CONTROLLER.set(1)
    tk.Checkbutton(input_panel, text="Export controllers", variable=EXPORT_CONTROLLER).grid(row=3,
                                                                                            column=4,
                                                                                            sticky=tk.W)
    params.export_controller = EXPORT_CONTROLLER.get

    CV_VERIFY = tk.IntVar()
    CV_VERIFY.set(1)
    tk.Checkbutton(input_panel, text="Verify conversion", variable=CV_VERIFY).grid(row=4, column=4,
                                                                                   sticky=tk.W)
    params.is_to_verify = CV_VERIFY.get

    RUN_DIAGNOSTIC = tk.IntVar()
    RUN_DIAGNOSTIC.set(0)
    tk.Checkbutton(input_panel, text="Diagnostic report", variable=RUN_DIAGNOSTIC).grid(row=5,
                                                                                        column=4,
                                                                                        sticky=tk.W)
    params.is_to_diagnostic = RUN_DIAGNOSTIC.get

    LOGGER_DEBUG = tk.IntVar()
    tk.Checkbutton(input_panel, text="Logger in debug mode", variable=LOGGER_DEBUG).grid(row=6,
                                                                                         column=4,
                                                                                         sticky=tk.W)
    params.is_debug = LOGGER_DEBUG.get

    CREATE_SECTIONS = tk.IntVar()
    CREATE_SECTIONS.set(0)
    tk.Checkbutton(input_panel, text="Create line sections as separate lines", variable=CREATE_SECTIONS).grid(row=7,
                                                                                         column=4,
                                                                                         sticky=tk.W)
    params.create_sections_entry = CREATE_SECTIONS.get

    REPLACE_ZERO_BRANCHES = tk.IntVar()
    REPLACE_ZERO_BRANCHES.set(0)
    tk.Checkbutton(input_panel, text="Replace low-impedance branches with switches",
                   variable=REPLACE_ZERO_BRANCHES).grid(row=8, column=4, sticky=tk.W)
    params.replace_zero_branches = REPLACE_ZERO_BRANCHES.get

    tk.Label(input_panel, anchor='w', text='Cut-off Z:').grid(row=8, column=5, sticky=tk.E, pady=0)
    tk.Label(input_panel, anchor='w', text='Ohm').grid(row=8, column=7, sticky=tk.W, pady=0)

    min_ohm = tk.Entry(input_panel, width=14)
    min_ohm.delete(0, tk.END)
    min_ohm.grid(row=8, column=6, padx=2, pady=2, sticky=tk.W)
    min_ohm.insert(0, "0.01")
    # entry_fname.insert(0, 'test') ##for testing
    params.min_ohm_entry = min_ohm

    REPLACE_INF_BRANCHES = tk.IntVar()
    REPLACE_INF_BRANCHES.set(0)
    tk.Checkbutton(input_panel, text="Deactivate high-impedance branches",
                   variable=REPLACE_INF_BRANCHES).grid(row=9, column=4, sticky=tk.W)
    params.replace_inf_branches = REPLACE_INF_BRANCHES.get

    tk.Label(input_panel, anchor='w', text='Cut-off Z:').grid(row=9, column=5, sticky=tk.E, pady=0)
    tk.Label(input_panel, anchor='w', text='Ohm').grid(row=9, column=7, sticky=tk.W, pady=0)

    max_ohm = tk.Entry(input_panel, width=14)
    max_ohm.delete(0, tk.END)
    max_ohm.grid(row=9, column=6, padx=2, pady=2, sticky=tk.W)
    max_ohm.insert(0, "1000")
    # entry_fname.insert(0, 'test') ##for testing
    params.max_ohm_entry = max_ohm
    
    # row 2 col 4
    stop_button = tk.Button(input_panel, text='Cancel', width=8,
                            command=lambda: cancel(app, input_panel))
    stop_button.grid(row=2, column=7, sticky='e', padx=4, pady=4)

    # app, get_dst_dir, input_panel, entry_fname, is_to_verify, is_debug, pv_as_slack,
    # pf_variable_p_loads, pf_variable_p_gen, flag_graphics
    start_button = tk.Button(input_panel, text='Export', width=8,
                             command=lambda: calc(app, **params))
    start_button.grid(row=1, column=7, sticky='e', padx=4, pady=0)

    input_panel.mainloop()

# if __name__ == '__main__':
#     import powerfactory
#
#     app = powerfactory.GetApplication()
#     app.Show()
#
#     logger, app_handler = logger_setup.setup_logger(app, 'INFO')
#     make_gui(app, 'test', browse_dst_test, calc_test)
# else:
#     logger = logging.getLogger(__name__)
#     # logger.setLevel('DEBUG')
